-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("announce", "AceEvent-3.0")

-----------------------
--- Library loading ---
-----------------------
local LA = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-Announce")

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local format = string.format
local GetNumPartyMembers, SendChatMessage = GetNumPartyMembers, SendChatMessage

-----------------
--- variables ---
-----------------
module.modName = LA["Quests Announcement"]
local prdb, options

-----------------------
--- Default profile ---
-----------------------
local defaults = {
	profile = {
		playerObjectiveProgress = true,
		playerObjectiveComplete = true,
		playerQuestComplete = true,
		playerQuestFailed = true,
		objectiveNotificationWhenComplete = false,
	},
}

---------------
--- Options ---
---------------
local optGetter, optSetter
do
	function optGetter(info)
		return prdb[info[#info]]
	end
	function optSetter(info, value)
		prdb[info[#info]] = value
	end
end

local function getOptions()
	if not options then
		options = {
			type = "group",
			name = module.modName,
			desc = module.modName,
			get = optGetter,
			set = optSetter,
			args = {
				playerObjectiveProgress = {
					name = LA["Quests objectives progress annoucement"],
					type = "toggle",
					width = "full",
					order = 12,
				},
				playerObjectiveComplete = {
					name = LA["Quests objectives complete announcement"],
					type = "toggle",
					width = "full",
					order = 13,
				},
				playerQuestComplete = {
					name = LA["Quests complete announcement"],
					type = "toggle",
					width = "full",
					order = 14,
				},
				playerQuestFailed = {
					name = LA["Quests failed announcement"],
					type = "toggle",
					width = "full",
					order = 15,
				},
			},
		}
	end
	return options
end

----------------------
--- Initialization ---
----------------------
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("announce", defaults, "Default")
	prdb = self.db.profile
	self:SetEnabledState(core:GetModuleEnabled("announce"))
	core:RegisterModuleOptions("announce", getOptions, module.modName)
end

function module:OnEnable()
	core.RegisterCallback(self, "playerQuestObjectiveProgress")
	core.RegisterCallback(self, "playerQuestComplete")
	core.RegisterCallback(self, "playerQuestFailed")
end

function module:OnDisable()
	core.UnregisterAllCallbacks(self)
	self:Refresh()
end

function module:Refresh()
	prdb = self.db.profile
end

---------------------------
--- Main Module Program ---
---------------------------
function module:playerQuestObjectiveProgress(event, firstRun, commHandler, uid, objective, got, nGot)
	if nGot == core.questsObjectives[uid][objective].nNeed and prdb.playerObjectiveComplete and not firstRun and GetNumPartyMembers() > 0 then
		if not prdb.objectiveNotificationWhenComplete and core.quests[uid].complete[core.playerName] ~= 0 then return end
		SendChatMessage(format(LA["(%s) %s : finished"], core.quests[uid].title, core.questsObjectives[uid][objective].desc), "PARTY")
	elseif prdb.playerObjectiveProgress and not firstRun and nGot > 0 and GetNumPartyMembers() > 0 then
		SendChatMessage(format("(%s) %s : %s/%s", core.quests[uid].title, core.questsObjectives[uid][objective].desc , got, core.questsObjectives[uid][objective].need), "PARTY")
	end
end

function module:playerQuestComplete(event, firstRun, commHandler, uid)
	if prdb.playerQuestComplete and not firstRun and GetNumPartyMembers() > 0 then
		SendChatMessage(format(LA["(%s) : finished"], core.quests[uid].title), "PARTY")
	end
end

function module:playerQuestFailed(event, firstRun, commHandler, uid)
	if prdb.playerQuestFailed and not firstRun and GetNumPartyMembers() > 0 then
		SendChatMessage(format(LA["(%s) : failed"], core.quests[uid].title), "PARTY")
	end
end