---------------------------
--- Core Addon Creation ---
---------------------------
local tomQuest2 = LibStub("AceAddon-3.0"):NewAddon("tomQuest2", "AceEvent-3.0", "AceTimer-3.0", "AceSerializer-3.0", "AceHook-3.0")
tomQuest2.cbh = LibStub("CallbackHandler-1.0"):New(tomQuest2)

-----------------------
--- Library loading ---
-----------------------
local LC = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-Core")
local BC = LibStub("LibBabble-Class-3.0"):GetLookupTable()
local LibTourist = LibStub("LibTourist-3.0")

--------------------------------------
--- Copy global functions to local ---
--------------------------------------
local pairs, next, type = pairs, next, type
local format, lower, upper, match, gsub, strtrim, find = string.format, string.lower, string.upper, string.match, string.gsub, strtrim, string.find
local tonumber = tonumber
local GetTime = GetTime
local GetDailyQuestsCompleted = GetDailyQuestsCompleted
local GetQuestGreenRange, QuestDifficultyColor = GetQuestGreenRange, QuestDifficultyColor
local GetQuestLink = GetQuestLink
local GetQuestLogLeaderBoard, GetQuestLogQuestText, GetQuestLogPushable = GetQuestLogLeaderBoard, GetQuestLogQuestText, GetQuestLogPushable
local GetQuestLogSelection, SelectQuestLogEntry, QuestLog_SetSelection = GetQuestLogSelection, SelectQuestLogEntry, QuestLog_SetSelection
local GetQuestLogRewardTalents, GetQuestLogRewardMoney, GetQuestLogRewardHonor = GetQuestLogRewardTalents, GetQuestLogRewardMoney, GetQuestLogRewardHonor
local GetQuestLogRewardSpell, GetQuestLogRewardTitle, GetQuestLogRewardInfo = GetQuestLogRewardSpell, GetQuestLogRewardTitle, GetQuestLogRewardInfo
local GetQuestLogChoiceInfo, GetQuestLogItemLink, GetQuestLogTitle = GetQuestLogChoiceInfo, GetQuestLogItemLink, GetQuestLogTitle
local GetNumQuestLogEntries, GetNumQuestLeaderBoards = GetNumQuestLogEntries, GetNumQuestLeaderBoards
local GetNumQuestLogRewards, GetNumQuestLogChoices = GetNumQuestLogRewards, GetNumQuestLogChoices
local GetNumPartyMembers, InCombatLockdown = GetNumPartyMembers, InCombatLockdown
local GetQuestLogSpecialItemInfo, UseQuestLogSpecialItem = GetQuestLogSpecialItemInfo, UseQuestLogSpecialItem
local FauxScrollFrame_GetOffset, FauxScrollFrame_SetOffset = FauxScrollFrame_GetOffset, FauxScrollFrame_SetOffset
local QuestLogListScrollFrame, QuestLogListScrollFrameScrollBar = QuestLogListScrollFrame, QuestLogListScrollFrameScrollBar
local HideUIPanel, ShowUIPanel = HideUIPanel, ShowUIPanel
local UnitName, UnitClass, UnitLevel = UnitName, UnitClass, UnitLevel
local GetMacroIndexByName, CreateMacro, EditMacro, PickupMacro = GetMacroIndexByName, CreateMacro, EditMacro, PickupMacro

-----------------
--- variables ---
-----------------
local _
local prdb
local dbVersion, lastCompatibleDB = 2, 1
tomQuest2.playerName = UnitName("player")
_, tomQuest2.playerClass = UnitClass("player")
tomQuest2.playerLevel = UnitLevel("player")
local localClass = BC[lower(tomQuest2.playerClass):gsub("^%l", upper)]
local questLogUpdatePending = false
local questLogNeedUpdate = false
local firstRun = true
local objectsPattern = '^' .. QUEST_OBJECTS_FOUND:gsub('(%%%d?$?.)', '(.+)') .. '$' --QUEST_OBJECTS_FOUND = "%s: %d/%d"
local monstersPattern = '^' .. QUEST_MONSTERS_KILLED:gsub('(%%%d?$?.)', '(.+)') .. '$' --QUEST_MONSTERS_KILLED = "%s slain: %d/%d"
local factionPattern = '^' .. QUEST_FACTION_NEEDED:gsub('(%%%d?$?.)', '(.+)') .. '$' --QUEST_FACTION_NEEDED = "%s: %s / %s"
local playersPattern = '^'.. QUEST_PLAYERS_KILLED:gsub('(%%%d?$?.)', '(.+)') .. '$' --QUEST_PLAYERS_KILLED
local DUNGEON = LFG_TYPE_DUNGEON
local HEROIC = DUNGEON_DIFFICULTY2
local collapsedHeader = {}
local abandoning = true
local currentUidMacro = nil
local reactions = {
	[FACTION_STANDING_LABEL1] = 1,
	[FACTION_STANDING_LABEL1_FEMALE] = 1,
	[FACTION_STANDING_LABEL2] = 2,
	[FACTION_STANDING_LABEL2_FEMALE] = 2,
	[FACTION_STANDING_LABEL3] = 3,
	[FACTION_STANDING_LABEL3_FEMALE] = 3,
	[FACTION_STANDING_LABEL4] = 4,
	[FACTION_STANDING_LABEL4_FEMALE] = 4,
	[FACTION_STANDING_LABEL5] = 5,
	[FACTION_STANDING_LABEL5_FEMALE] = 5,
	[FACTION_STANDING_LABEL6] = 6,
	[FACTION_STANDING_LABEL6_FEMALE] = 6,
	[FACTION_STANDING_LABEL7] = 7,
	[FACTION_STANDING_LABEL7_FEMALE] = 7,
	[FACTION_STANDING_LABEL8] = 8,
	[FACTION_STANDING_LABEL8_FEMALE] = 8,
}

local tags = {[ELITE] = "elite", [GROUP] = "group", [PVP] = "pvp", [RAID] = "raid", [DUNGEON] = "dungeon", [HEROIC] = "heroic", [DAILY] = "daily"}
tomQuest2.numCompletedQuests = 0
tomQuest2.numCompletedDailyQuests = 0
tomQuest2.numEntries, tomQuest2.numQuests = 0, 0
tomQuest2.numPartyMembers = 0
tomQuest2.party = {}
tomQuest2.quests = {}
tomQuest2.questsObjectives = {}
tomQuest2.questsItems = {}
tomQuest2.questsMobs = {}
tomQuest2.questsMobsItems = {}
tomQuest2.questsByZone = {}
tomQuest2.questsIdToUid = {}
tomQuest2.questsTitleToUid = {}
tomQuest2.numTQ_PlQC = 0
tomQuest2.numTQ_PaQC = 0
tomQuest2.numTQ_PlQC_BML = 0
tomQuest2.numTQ_PaQC_BML = 0

----------------------
-- Table recycling ---
----------------------
local tablePool = {}
setmetatable(tablePool, {__mode = "kv"})

local function new()
	local t = next(tablePool) or {}
	tablePool[t] = nil
	return t
end

local function del(t)
	if type(t) == "table" then
		for k, v in pairs(t) do
			if type(v) == "table" then
				v = del(v)
			end
			t[k] = nil
		end
		t[true] = true
		t[true] = nil
		setmetatable(t, nil)
		tablePool[t] = true
	end
	return nil
end

-------------------
--- Class Color ---
-------------------
tomQuest2.classColors = {}
for class, color in pairs(CUSTOM_CLASS_COLORS or RAID_CLASS_COLORS) do
	tomQuest2.classColors[class] = format("%02x%02x%02x", color.r*255, color.g*255, color.b*255)
end
local playerColor = tomQuest2.classColors[tomQuest2.playerClass]

function tomQuest2:updateClassColors()
	for class, color in pairs(CUSTOM_CLASS_COLORS) do
		tomQuest2.classColors[class] = format("%02x%02x%02x", color.r*255, color.g*255, color.b*255)
	end
	if tomQuest2.classColors[tomQuest2.playerClass] ~= playerColor then
		playerColor = tomQuest2.classColors[tomQuest2.playerClass]
		for zone in pairs(tomQuest2.questsByZone) do
			if zone == localClass then
				tomQuest2.questsByZone[zone].color = playerColor
				tomQuest2.cbh:Fire("updateQuestsTooltip")
			end
		end
	end
	if tomQuest2.party then
		local colorUpdated = false
		for playerName, pData in pairs(tomQuest2.party) do
			if pData.color ~= tomQuest2.classColors[pData.class] then
				tomQuest2.party[playerName].color = tomQuest2.classColors[pData.class]
				colorUpdated = true
			end
		end
		if colorUpdated == true then
			for uid, qData in pairs(tomQuest2.quests) do
				tomQuest2:computeQuestStatus(uid)
				if qData.numObjectives > 0 then
					for objective in pairs(tomQuest2.questsObjectives[uid]) do
						tomQuest2:computeQuestObjectiveStatus(uid, objective)
					end
				end
			end
			tomQuest2.cbh:Fire("updateQuestsTooltip")
		end
	end
end

-----------------------
--- Default profile ---
-----------------------
local defaults = {
	profile = {
		modules = {
			dailyTooltip = true,
			notifications = true,
			informations = true,
			questsTooltip = true,
			questsTracker = true,
			soundNotifications = true,
			syncParty = true,
			tagQuestGivers = true,
			tagQuestLog = true,
			tooltips = true,
			announce = false,
			achievementTracker = true,
			questsInfo = true,
		},
		quests = {
			hideDefaultBlizTracker = true,
			maxLevelDiff = 3,
			eliteAddLevel = 3,
			groupAddLevel = 1,
			dungeonAddLevel = 2,
			heroicAddLevel = 3,
			elite = '+',
			group = 'g',
			pvp = 'p',
			raid = 'r',
			dungeon = 'd',
			heroic = 'd+',
			daily = "\226\128\162",
			rj0 = 1,
			gj0 = 0,
			bj0 = 0,
			ri1 = 0,
			rj1 = 1,
			gi1 = 2,
			gj1 = 0,
			bi1 = 0,
			bj1 = 0,
			ri2 = -2,
			rj2 = 2,
			gi2 = 0,
			gj2 = 1,
			bi2 = 0,
			bj2 = 0,
			ri3 = -2,
			rj3 = 2,
			gi3 = 0,
			gj3 = 1,
			bi3 = 0,
			bj3 = 0,
		},
	},
}

----------------------
--- Initialization ---
----------------------
tomQuest2:SetDefaultModuleState(false)
function tomQuest2:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("tomQuest2DB", defaults, "Default")
	self.db.RegisterCallback(self, "OnProfileChanged", "Refresh")
	self.db.RegisterCallback(self, "OnProfileCopied", "Refresh")
	self.db.RegisterCallback(self, "OnProfileReset", "Refresh")
	prdb = self.db.profile

	if prdb.version and prdb.version < lastCompatibleDB then
		self.db:ResetDB()
		tomQuest2:Refresh()
	end
	prdb.version = dbVersion
	self:SetupOptions()

	self:RegisterEvent("QUEST_LOG_UPDATE")
	self:RegisterEvent("PLAYER_LEVEL_UP")
end

function tomQuest2:OnEnable()
	if CUSTOM_CLASS_COLORS then
		CUSTOM_CLASS_COLORS:RegisterCallback("updateClassColors", tomQuest2)
	end
	self:SecureHook("AbandonQuest")
	self:SecureHook("WatchFrame_Update")
	WATCHFRAME_MAXQUESTS = 100
	WATCHFRAME_MAXACHIEVEMENTS = 100
	WATCHFRAME_LINEHEIGHT = 1
	MAX_WATCHABLE_QUESTS = 100
end

function tomQuest2:Refresh()
	prdb = self.db.profile
	prdb.version = dbVersion
	for k, v in self:IterateModules() do
		if self:GetModuleEnabled(k) and not v:IsEnabled() then
			self:EnableModule(k)
		elseif not self:GetModuleEnabled(k) and v:IsEnabled() then
			self:DisableModule(k)
		end
		if v.Refresh and type(v.Refresh) == "function" then
			v:Refresh()
		end
	end
	tomQuest2:computeQuestTag()
	tomQuest2:computeQuestsObjectivesStatus()
end

function tomQuest2:GetModuleEnabled(module)
	return prdb.modules[module]
end

function tomQuest2:SetModuleEnabled(module, value)
	local old = prdb.modules[module]
	prdb.modules[module] = value
	if old ~= value then
		if value then
			self:EnableModule(module)
		else
			self:DisableModule(module)
		end
	end
end

------------------------------------
--- Bliz Quest Tracker show hide ---
------------------------------------
function tomQuest2:WatchFrame_Update()
	if prdb.quests.hideDefaultBlizTracker then
		WatchFrame:Hide()
		WatchFrameTitleButton:Hide()
		WatchFrameLines:Hide()
	else
		WatchFrame:Show()
		WatchFrameTitleButton:Show()
		WatchFrameLines:Show()
	end
end

function tomQuest2:hideShowDefaultBlizTracker()
	if prdb.quests.hideDefaultBlizTracker then
		WatchFrame:Hide()
		WatchFrameTitleButton:Hide()
		WatchFrameLines:Hide()
	else
		WatchFrame:Show()
		WatchFrameTitleButton:Show()
		WatchFrameLines:Show()
	end
end

-------------------------
--- Hook AbandonQuest ---
-------------------------
function tomQuest2:AbandonQuest()
	abandoning = true
end

-----------------------
--- Tools Functions ---
-----------------------
function tomQuest2:GetDifficultyColor(levelDiff)
	local color
	if ( levelDiff >= 5 ) then
		color = QuestDifficultyColor["impossible"]
	elseif ( levelDiff >= 3 ) then
		color = QuestDifficultyColor["verydifficult"]
	elseif ( levelDiff >= -2 ) then
		color = QuestDifficultyColor["difficult"]
	elseif ( -levelDiff <= GetQuestGreenRange() ) then
		color = QuestDifficultyColor["standard"]
	else
		color = QuestDifficultyColor["trivial"]
	end
	return color;
end

function tomQuest2:getColorFromCompletionRGB(percent)
	local r, g, b
	if percent == 0 then
		r, g, b = prdb.quests.rj0, prdb.quests.gj0, prdb.quests.bj0
		if r > 1 then r = 1 elseif r < 0 then r = 0 end
		if g > 1 then g = 1 elseif g < 0 then g = 0 end
		if b > 1 then b = 1 elseif b < 0 then b = 0 end
	elseif percent == 1 then
		r, g, b = prdb.quests.ri3 + prdb.quests.rj3, prdb.quests.gi3 + prdb.quests.gj3, prdb.quests.bi3 + prdb.quests.bj3
		if r > 1 then r = 1 elseif r < 0 then r = 0 end
		if g > 1 then g = 1 elseif g < 0 then g = 0 end
		if b > 1 then b = 1 elseif b < 0 then b = 0 end
	elseif percent <= 0.5 then
		r, g, b = prdb.quests.ri1*percent + prdb.quests.rj1, prdb.quests.gi1*percent + prdb.quests.gj1, prdb.quests.bi1*percent + prdb.quests.bj1
		if r > 1 then r = 1 elseif r < 0 then r = 0 end
		if g > 1 then g = 1 elseif g < 0 then g = 0 end
		if b > 1 then b = 1 elseif b < 0 then b = 0 end
	else
		r, g, b = prdb.quests.ri2*percent + prdb.quests.rj2, prdb.quests.gi2*percent + prdb.quests.gj2, prdb.quests.bi2*percent + prdb.quests.bj2
		if r > 1 then r = 1 elseif r < 0 then r = 0 end
		if g > 1 then g = 1 elseif g < 0 then g = 0 end
		if b > 1 then b = 1 elseif b < 0 then b = 0 end
	end
	return r, g, b
end

function tomQuest2:getColorFromCompletionHEX(percent)
	local r, g, b
	if percent == 0 then
		r, g, b = prdb.quests.rj0, prdb.quests.gj0, prdb.quests.bj0
		if r > 1 then r = 1 elseif r < 0 then r = 0 end
		if g > 1 then g = 1 elseif g < 0 then g = 0 end
		if b > 1 then b = 1 elseif b < 0 then b = 0 end
	elseif percent == 1 then
		r, g, b = prdb.quests.ri3 + prdb.quests.rj3, prdb.quests.gi3 + prdb.quests.gj3, prdb.quests.bi3 + prdb.quests.bj3
		if r > 1 then r = 1 elseif r < 0 then r = 0 end
		if g > 1 then g = 1 elseif g < 0 then g = 0 end
		if b > 1 then b = 1 elseif b < 0 then b = 0 end
	elseif percent <= 0.5 then
		r, g, b = prdb.quests.ri1*percent + prdb.quests.rj1, prdb.quests.gi1*percent + prdb.quests.gj1, prdb.quests.bi1*percent + prdb.quests.bj1
		if r > 1 then r = 1 elseif r < 0 then r = 0 end
		if g > 1 then g = 1 elseif g < 0 then g = 0 end
		if b > 1 then b = 1 elseif b < 0 then b = 0 end
	else
		r, g, b = prdb.quests.ri2*percent + prdb.quests.rj2, prdb.quests.gi2*percent + prdb.quests.gj2, prdb.quests.bi2*percent + prdb.quests.bj2
		if r > 1 then r = 1 elseif r < 0 then r = 0 end
		if g > 1 then g = 1 elseif g < 0 then g = 0 end
		if b > 1 then b = 1 elseif b < 0 then b = 0 end
	end
	return format("%02x%02x%02x", 255*r, 255*g, 255*b)
end

--------------------------
--- Status Computation ---
--------------------------
function tomQuest2:computeQuestStatus(uid)
	tomQuest2.quests[uid].pqStatus = ""
	local partyComplete = tomQuest2.quests[uid].partyComplete
	local playerComplete = tomQuest2.quests[uid].playerComplete
	local totalCompleted = 0
	local numPlayers = 0
	for partyName, complete in pairs(tomQuest2.quests[uid].complete) do
		if partyName ~= tomQuest2.playerName and (complete == 1 or complete == -1) then
			numPlayers = numPlayers + 1
			totalCompleted = totalCompleted + complete
			if numPlayers > 1 then
				tomQuest2.quests[uid].pqStatus = format("%s, |cff%s%s|r: %s",tomQuest2.quests[uid].pqStatus, tomQuest2.party[partyName].color or "ffffff", partyName:sub(1, 3), complete == 1 and LC["Finished"] or LC["Failed"])
			else
				tomQuest2.quests[uid].pqStatus = format("|cff%s%s|r: %s", tomQuest2.party[partyName].color or "ffffff", partyName:sub(1, 3), complete == 1 and LC["Finished"] or LC["Failed"])
			end
		end
	end
	if tomQuest2.quests[uid].complete[tomQuest2.playerName] and (tomQuest2.quests[uid].complete[tomQuest2.playerName] == 1 or tomQuest2.quests[uid].complete[tomQuest2.playerName] == -1) then
		numPlayers = numPlayers + 1
		totalCompleted = totalCompleted + tomQuest2.quests[uid].complete[tomQuest2.playerName]
		if tomQuest2.quests[uid].complete[tomQuest2.playerName] == 1 then
			tomQuest2.quests[uid].playerComplete = 1
		else
			tomQuest2.quests[uid].playerComplete = 0
		end
		if numPlayers > 1 then
			tomQuest2.quests[uid].pqStatus = format("%s, %s", tomQuest2.quests[uid].pqStatus, tomQuest2.quests[uid].complete[tomQuest2.playerName] == 1 and LC["Finished"] or LC["Failed"])
		else
			tomQuest2.quests[uid].pqStatus = format("%s", tomQuest2.quests[uid].complete[tomQuest2.playerName] == 1 and LC["Finished"] or LC["Failed"])
		end
	end
	if totalCompleted == tomQuest2.quests[uid].playersOnQuest then
		tomQuest2.quests[uid].partyComplete = 1
	else
		tomQuest2.quests[uid].partyComplete = 0
	end
	if tomQuest2.quests[uid].tracked == 1 then
		if tomQuest2.quests[uid].complete[tomQuest2.playerName] then
			tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ_PlQC = tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ_PlQC + tomQuest2.quests[uid].playerComplete - playerComplete
			tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ_PaQC = tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ_PaQC + tomQuest2.quests[uid].partyComplete - partyComplete
			tomQuest2.numTQ_PlQC = tomQuest2.numTQ_PlQC + tomQuest2.quests[uid].playerComplete - playerComplete
			tomQuest2.numTQ_PaQC = tomQuest2.numTQ_PaQC + tomQuest2.quests[uid].partyComplete - partyComplete
			if tomQuest2.quests[uid].weightedLevel - tomQuest2.playerLevel <= tonumber(prdb.quests.maxLevelDiff) then
				tomQuest2.numTQ_PlQC_BML = tomQuest2.numTQ_PlQC_BML + tomQuest2.quests[uid].playerComplete - playerComplete
				tomQuest2.numTQ_PaQC_BML = tomQuest2.numTQ_PaQC_BML + tomQuest2.quests[uid].partyComplete - partyComplete
			end
		else
			tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numTQ_PaQC = tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numTQ_PaQC + tomQuest2.quests[uid].partyComplete - partyComplete
			tomQuest2.numTQ_PaQC = tomQuest2.numTQ_PaQC + tomQuest2.quests[uid].partyComplete - partyComplete
			if tomQuest2.quests[uid].weightedLevel - tomQuest2.playerLevel <= tonumber(prdb.quests.maxLevelDiff) then
				tomQuest2.numTQ_PaQC_BML = tomQuest2.numTQ_PaQC_BML + tomQuest2.quests[uid].partyComplete - partyComplete
			end
		end
	end
end

function tomQuest2:computeQuestObjectiveStatus(uid, objective)
	tomQuest2.questsObjectives[uid][objective].qoStatus = ""
	local totalGot = 0
	local numPlayers = 0
	local pGot, pNGot
	for partyName, qoData in pairs(tomQuest2.questsObjectives[uid][objective].got) do
		local got, nGot= qoData:match("^(.-)~(%d-)$")
		nGot = tonumber(nGot)
		if partyName ~= tomQuest2.playerName then
			numPlayers = numPlayers + 1
			totalGot = totalGot + nGot
			if numPlayers > 1 then
				tomQuest2.questsObjectives[uid][objective].qoStatus = format("%s, |cff%s%s|r: |cff%s%s|r",tomQuest2.questsObjectives[uid][objective].qoStatus, tomQuest2.party[partyName].color, partyName:sub(1,3), tomQuest2:getColorFromCompletionHEX(nGot / tomQuest2.questsObjectives[uid][objective].nNeed), nGot == tomQuest2.questsObjectives[uid][objective].nNeed and LC["d"] or got)
				tomQuest2.questsObjectives[uid][objective].qoStatusNC = format("%s, |cff%s%s|r: %s",tomQuest2.questsObjectives[uid][objective].qoStatusNC, tomQuest2.party[partyName].color, partyName:sub(1,3), nGot == tomQuest2.questsObjectives[uid][objective].nNeed and LC["d"] or got)
			else
				tomQuest2.questsObjectives[uid][objective].qoStatus = format("|cff%s%s|r: |cff%s%s|r", tomQuest2.party[partyName].color, partyName:sub(1,3), tomQuest2:getColorFromCompletionHEX(nGot / tomQuest2.questsObjectives[uid][objective].nNeed), nGot == tomQuest2.questsObjectives[uid][objective].nNeed and LC["d"] or got)
				tomQuest2.questsObjectives[uid][objective].qoStatusNC = format("|cff%s%s|r: %s", tomQuest2.party[partyName].color, partyName:sub(1,3), nGot == tomQuest2.questsObjectives[uid][objective].nNeed and LC["d"] or got)
			end
		else
			pGot, pNGot = got, nGot
		end
	end
	if pNGot then
		numPlayers = numPlayers + 1
		totalGot = totalGot + pNGot
		if numPlayers > 1 then
			tomQuest2.questsObjectives[uid][objective].qoStatus = format("%s, |cff%s%s|r", tomQuest2.questsObjectives[uid][objective].qoStatus, tomQuest2:getColorFromCompletionHEX(pNGot/tomQuest2.questsObjectives[uid][objective].nNeed), pNGot == tomQuest2.questsObjectives[uid][objective].nNeed and LC["d"] or pGot)
			tomQuest2.questsObjectives[uid][objective].qoStatusNC = format("%s, %s", tomQuest2.questsObjectives[uid][objective].qoStatusNC, pNGot == tomQuest2.questsObjectives[uid][objective].nNeed and LC["d"] or pGot)
		else
			if not tomQuest2.questsObjectives[uid][objective].nNeed then
				DEFAULT_CHAT_FRAME:AddMessage("error in objective informations please report this to Gagou on WowAce, WowInterface, or directly on EU Drek'Thar")
				DEFAULT_CHAT_FRAME:AddMessage("uid"..uid)
				DEFAULT_CHAT_FRAME:AddMessage("objective :"..objective)
				DEFAULT_CHAT_FRAME:AddMessage("objective :"..objective)
				DEFAULT_CHAT_FRAME:AddMessage("tomQuest2.questsObjectives[uid][objective].qType :"..tomQuest2.questsObjectives[uid][objective].qType)
				DEFAULT_CHAT_FRAME:AddMessage("tomQuest2.questsObjectives[uid][objective].desc :"..tomQuest2.questsObjectives[uid][objective].desc)
			end
			tomQuest2.questsObjectives[uid][objective].qoStatus = format("|cff%s%s|r", tomQuest2:getColorFromCompletionHEX(pNGot/tomQuest2.questsObjectives[uid][objective].nNeed), pNGot == tomQuest2.questsObjectives[uid][objective].nNeed and LC["d"] or pGot)
			tomQuest2.questsObjectives[uid][objective].qoStatusNC = pNGot == tomQuest2.questsObjectives[uid][objective].nNeed and LC["d"] or pGot
		end
		if pNGot == tomQuest2.questsObjectives[uid][objective].nNeed then
			tomQuest2.questsObjectives[uid][objective].playerComplete = 1
		else
			tomQuest2.questsObjectives[uid][objective].playerComplete = 0
		end
	end
	if totalGot == tomQuest2.quests[uid].playersOnQuest * tomQuest2.questsObjectives[uid][objective].nNeed then
		tomQuest2.questsObjectives[uid][objective].partyComplete = 1
	else
		tomQuest2.questsObjectives[uid][objective].partyComplete = 0
	end
	if not tomQuest2.questsObjectives[uid][objective].nNeed then
		DEFAULT_CHAT_FRAME:AddMessage("error in objective informations please report this to Gagou on WowAce, WowInterface, or directly on EU Drek'Thar")
		DEFAULT_CHAT_FRAME:AddMessage("uid"..uid)
		DEFAULT_CHAT_FRAME:AddMessage("objective :"..objective)
		DEFAULT_CHAT_FRAME:AddMessage("objective :"..objective)
		DEFAULT_CHAT_FRAME:AddMessage("tomQuest2.questsObjectives[uid][objective].qType :"..tomQuest2.questsObjectives[uid][objective].qType)
		DEFAULT_CHAT_FRAME:AddMessage("tomQuest2.questsObjectives[uid][objective].desc :"..tomQuest2.questsObjectives[uid][objective].desc)
	end
	tomQuest2.questsObjectives[uid][objective].qoColor = tomQuest2:getColorFromCompletionHEX(totalGot/ (tomQuest2.quests[uid].playersOnQuest * tomQuest2.questsObjectives[uid][objective].nNeed))
	tomQuest2.questsObjectives[uid][objective].qoStatus = format("%s/|cff%s%s|r", tomQuest2.questsObjectives[uid][objective].qoStatus, tomQuest2.questsObjectives[uid][objective].qoColor, tomQuest2.questsObjectives[uid][objective].need)
	tomQuest2.questsObjectives[uid][objective].qoStatusNC = format("%s/%s", tomQuest2.questsObjectives[uid][objective].qoStatusNC, tomQuest2.questsObjectives[uid][objective].need)
end

function tomQuest2:computeQuestsObjectivesStatus()
	for uid, qoData in pairs (tomQuest2.questsObjectives) do
		for objective in pairs (qoData) do
			tomQuest2:computeQuestObjectiveStatus(uid, objective)
		end
	end
	tomQuest2.cbh:Fire("updateQuestsTracker")
	tomQuest2.cbh:Fire("questLogUpdated", true, true)
end

------------------------
--- Quests Functions ---
------------------------
function tomQuest2:addZone(zone)
	tomQuest2.questsByZone[zone] = new()
	if zone == LC["PARTYQUESTSZONE"] then
		tomQuest2.questsByZone[zone].minLevel, tomQuest2.questsByZone[zone].maxLevel, tomQuest2.questsByZone[zone].level = "", "", ""
		tomQuest2.questsByZone[zone].color = "ffffff"
		tomQuest2.questsByZone[zone].zone = true
	elseif zone == localClass then
		tomQuest2.questsByZone[zone].minLevel, tomQuest2.questsByZone[zone].maxLevel, tomQuest2.questsByZone[zone].level = "", "", ""
		tomQuest2.questsByZone[zone].color = playerColor
	else
		tomQuest2.questsByZone[zone].minLevel, tomQuest2.questsByZone[zone].maxLevel = LibTourist:GetLevel(zone)
		if tomQuest2.questsByZone[zone].minLevel == 0 then
			tomQuest2.questsByZone[zone].level = ""
			tomQuest2.questsByZone[zone].color = "ffffff"
		else
			tomQuest2.questsByZone[zone].level = format("(%s-%s)", tomQuest2.questsByZone[zone].minLevel, tomQuest2.questsByZone[zone].maxLevel)
			local r, g, b = LibTourist:GetLevelColor(zone)
			tomQuest2.questsByZone[zone].color = format("%02x%02x%02x", r*255, g*255, b*255)
			tomQuest2.questsByZone[zone].zone = true
		end
	end
	tomQuest2.questsByZone[zone].numQuests = 0
	tomQuest2.questsByZone[zone].numTQ = 0
	tomQuest2.questsByZone[zone].numTQ_PlQC = 0
	tomQuest2.questsByZone[zone].numTQ_PaQC = 0
	tomQuest2.questsByZone[zone].numTQ_BML = 0
	tomQuest2.questsByZone[zone].numTQ_PlQC_BML = 0
	tomQuest2.questsByZone[zone].numTQ_PaQC_BML = 0
	tomQuest2.questsByZone[zone].quests = new()
end

function tomQuest2:addQuest(sender, first, uid, title, level, tag, group, daily, numObjectives, complete, zone, id)
	if not tomQuest2.quests[uid] then
		tomQuest2.quests[uid] = new()
		tomQuest2.quests[uid].playersOnQuest = 0
		tomQuest2.quests[uid].playerComplete = 0
		tomQuest2.quests[uid].partyComplete = 0
		tomQuest2.quests[uid].tracked = 0
		tomQuest2.quests[uid].title = title
		tomQuest2.quests[uid].level = level
		tomQuest2.quests[uid].zone = zone
		tomQuest2.quests[uid].numObjectives = numObjectives
		tomQuest2.quests[uid].complete = new()
		tomQuest2.quests[uid].complete[sender] = complete
		local qc = tomQuest2:GetDifficultyColor(level - tomQuest2.playerLevel)
		tomQuest2.quests[uid].color = format("%02x%02x%02x", qc.r*255, qc.g*255, qc.b*255)
		tomQuest2.quests[uid].shortTag = tag
		tomQuest2.quests[uid].group = group
		tomQuest2.quests[uid].daily = daily
		if level > 0 or tag ~= "" or group > 0 or daily == 1 then
			tomQuest2.quests[uid].tag = format("[%s%s%s%s]", level or "", tag ~= "" and prdb.quests[tags[tag]] or "", group > 0 and group or "", daily == 1 and (tag ~= DAILY and prdb.quests[tags[DAILY]] or "") or "")
		else
			tomQuest2.quests[uid].tag = ""
		end
		if tag == ELITE and (prdb.quests.eliteAddLevel or 0) > 0 then
			tomQuest2.quests[uid].weightedLevel = (level or 0) + prdb.quests.eliteAddLevel
		elseif tag == GROUP and (prdb.quests.groupAddLevel or 0) > 0 then
			tomQuest2.quests[uid].weightedLevel = (level or 0) + (prdb.quests.groupAddLevel * (tomQuest2.quests[uid].group - (tomQuest2.numPartyMembers or (GetNumPartyMembers() or 0))))
		elseif tag == HEROIC and (prdb.quests.heroicAddLevel or 0) > 0 then
			tomQuest2.quests[uid].weightedLevel = (level or 0) + (prdb.quests.heroicAddLevel * (5 - (tomQuest2.numPartyMembers or (GetNumPartyMembers() or 0))))
		elseif tag == DUNGEON and (prdb.quests.dungeonAddLevel or 0) > 0 then
			tomQuest2.quests[uid].weightedLevel = (level or 0) + (prdb.quests.dungeonAddLevel * (5 - (tomQuest2.numPartyMembers or (GetNumPartyMembers() or 0))))
		else
			tomQuest2.quests[uid].weightedLevel = (level or 0)
		end
		if numObjectives > 0 then
			tomQuest2.questsObjectives[uid] = new()
		end
		if sender ~= tomQuest2.playerName then
			tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].quests[uid] = true
			tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numQuests = tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numQuests + 1
		else
			tomQuest2.questsByZone[zone].quests[uid] = true
			tomQuest2.questsByZone[zone].numQuests = tomQuest2.questsByZone[zone].numQuests + 1
			if zone == localClass then
				tomQuest2.quests[uid].class = true
			end
		end
	else
		tomQuest2.quests[uid].complete[sender] = complete
	end
	tomQuest2.quests[uid].playersOnQuest = tomQuest2.quests[uid].playersOnQuest + 1
	if sender == tomQuest2.playerName then
		if not tomQuest2.quests[uid].description or not tomQuest2.quests[uid].objective then
			SelectQuestLogEntry(id)
			tomQuest2.quests[uid].extras = true
			tomQuest2.quests[uid].specialItemLink, tomQuest2.quests[uid].specialItemIcon, tomQuest2.quests[uid].specialItemCharges = GetQuestLogSpecialItemInfo(id)
			tomQuest2.quests[uid].sharable = GetQuestLogPushable()
			tomQuest2.quests[uid].description, tomQuest2.quests[uid].objective = GetQuestLogQuestText()
			tomQuest2.quests[uid].money = GetQuestLogRewardMoney() or 0
			tomQuest2.quests[uid].honor = GetQuestLogRewardHonor() or 0
			tomQuest2.quests[uid].talent = GetQuestLogRewardTalents() or 0
			tomQuest2.quests[uid].titleReward = GetQuestLogRewardTitle() or ""
			tomQuest2.quests[uid].numRewards = GetNumQuestLogRewards() or 0
			tomQuest2.quests[uid].numChoices = GetNumQuestLogChoices() or 0
			tomQuest2.quests[uid].rewardSpellTexture, tomQuest2.quests[uid].rewardSpellName = GetQuestLogRewardSpell()
			tomQuest2.quests[uid].rewardSpellTexture = tomQuest2.quests[uid].rewardSpellTexture or ""
			tomQuest2.quests[uid].rewardSpellName = tomQuest2.quests[uid].rewardSpellName or ""
			if tomQuest2.quests[uid].numRewards > 0 then
				tomQuest2.quests[uid].itemRewards = new()
				for i = 1, tomQuest2.quests[uid].numRewards do
					tomQuest2.quests[uid].itemRewards[i] = new()
					tomQuest2.quests[uid].itemRewards[i].name, tomQuest2.quests[uid].itemRewards[i].texture, tomQuest2.quests[uid].itemRewards[i].numItems, tomQuest2.quests[uid].itemRewards[i].quality, tomQuest2.quests[uid].itemRewards[i].isUsable = GetQuestLogRewardInfo(i)
					tomQuest2.quests[uid].itemRewards[i].link = GetQuestLogItemLink('reward', i)
				end
			end
			if tomQuest2.quests[uid].numChoices > 0 then
				tomQuest2.quests[uid].itemChoices = new()
				for i = 1, tomQuest2.quests[uid].numChoices do
					tomQuest2.quests[uid].itemChoices[i] = new()
					tomQuest2.quests[uid].itemChoices[i].name, tomQuest2.quests[uid].itemChoices[i].texture, tomQuest2.quests[uid].itemChoices[i].numItems, tomQuest2.quests[uid].itemChoices[i].quality, tomQuest2.quests[uid].itemChoices[i].isUsable = GetQuestLogChoiceInfo(i)
					tomQuest2.quests[uid].itemChoices[i].link = GetQuestLogItemLink('choice', i)
				end
			end
		end
		if tomQuest2.quests[uid].playersOnQuest > 1 then
			tomQuest2.questsByZone[zone].quests[uid] = true
			tomQuest2.questsByZone[zone].numQuests = tomQuest2.questsByZone[zone].numQuests + 1
			tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].quests[uid] = del(tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].quests[uid])
			tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numQuests = tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numQuests - 1
			if tomQuest2.quests[uid].tracked == 1 then
				tomQuest2.questsByZone[zone].numTQ = tomQuest2.questsByZone[zone].numTQ + 1
				tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numTQ = tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numTQ - 1
			end
			if tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numQuests == 0 then
				tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]] = del(tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]])
			end
		end
		tomQuest2.cbh:Fire("addPlayerQuest", first, "q", uid, title, level, tag, group, daily, numObjectives, complete, zone)
	else
		tomQuest2.cbh:Fire("addPartyQuest", nil, nil, uid)
	end
	tomQuest2:computeQuestStatus(uid)
end

function tomQuest2:questComplete(sender, first, uid)
	tomQuest2.quests[uid].complete[sender] = 1
	tomQuest2:computeQuestStatus(uid)
	if sender == tomQuest2.playerName then
		tomQuest2.cbh:Fire("playerQuestComplete", first, "c", uid)
	else
		tomQuest2.cbh:Fire("partyQuestComplete", sender, uid)
	end
end

function tomQuest2:questFailed(sender, first, uid)
	tomQuest2.quests[uid].complete[sender] = -1
	tomQuest2:computeQuestStatus(uid)
	if sender == tomQuest2.playerName then
		tomQuest2.cbh:Fire("playerQuestFailed", first, "f", uid)
	else
		tomQuest2.cbh:Fire("partyQuestFailed", sender, uid)
	end
end

function tomQuest2:addQuestObjective(sender, first, uid, objective, desc, qType, got, need, nGot, nNeed)
	if not tomQuest2.questsObjectives[uid][objective] then
		tomQuest2.questsObjectives[uid][objective] = new()
		tomQuest2.questsObjectives[uid][objective].desc = desc
		tomQuest2.questsObjectives[uid][objective].qType = qType
		tomQuest2.questsObjectives[uid][objective].nNeed = nNeed
		tomQuest2.questsObjectives[uid][objective].need = need
		tomQuest2.questsObjectives[uid][objective].got = new()
		if qType == "object" then
			if not tomQuest2.questsItems[desc] then
				tomQuest2.questsItems[desc] = new()
				tomQuest2.questsItems[desc].numQuests = 0
				tomQuest2.questsItems[desc].quests = new()
			end
			if not tomQuest2.questsItems[desc].quests[uid] then
				tomQuest2.questsItems[desc].numQuests = tomQuest2.questsItems[desc].numQuests + 1
				tomQuest2.questsItems[desc].quests[uid] = objective
			end
			if MobMapDotParentFrame then
				if not MobMap_GetDropItemNameList then
					MobMap_LoadDatabase(7)
				end
				if MobMap_GetDropItemNameList then
					local mobMapDropList = MobMap_GetDropItemNameList(desc)
					if mobMapDropList then
						local mobMapDropChance = MobMap_GetDropChances(mobMapDropList[1])
						for k, v in pairs(mobMapDropChance) do
							if v.mobid and v.mobid ~= 0 then
								local mobName = MobMap_GetMobName(v.mobid)
								if not tomQuest2.questsItems[desc].mobDrop then
									tomQuest2.questsItems[desc].mobDrop = new()
								end
								tomQuest2.questsItems[desc].mobDrop[mobName] = true
								if not tomQuest2.questsMobsItems[mobName] then
									tomQuest2.questsMobsItems[mobName] = new()
									tomQuest2.questsMobsItems[mobName].items = new()
									tomQuest2.questsMobsItems[mobName].numItems = 0
									local zoneId = MobMap_GetMobZone(v.mobid)
									tomQuest2.questsMobsItems[mobName].mobLoc = new()
									local zoneName = "???"
									if zoneId.id ~= nil then
										zoneName = MobMap_GetZoneName(zoneId.id)
									end
									tomQuest2.questsMobsItems[mobName].mobLoc[zoneName] = true
								end
								if not tomQuest2.questsMobsItems[mobName].items[desc] then
									tomQuest2.questsMobsItems[mobName].numItems = tomQuest2.questsMobsItems[mobName].numItems + 1
									tomQuest2.questsMobsItems[mobName].items[desc] = new()
									tomQuest2.questsMobsItems[mobName].items[desc].numQuests = 0
									tomQuest2.questsMobsItems[mobName].items[desc].dropRate = v.chance
									tomQuest2.questsMobsItems[mobName].items[desc].quests = new()
								end
								if not tomQuest2.questsMobsItems[mobName].items[desc].quests[uid] then
									tomQuest2.questsMobsItems[mobName].items[desc].numQuests = tomQuest2.questsMobsItems[mobName].items[desc].numQuests + 1
									tomQuest2.questsMobsItems[mobName].items[desc].quests[uid] = objective
								end
							end
						end
					end
				end
			end
		elseif qType == "monster" then
			if not tomQuest2.questsMobs[desc] then
				tomQuest2.questsMobs[desc] = new()
				tomQuest2.questsMobs[desc].numQuests = 0
				tomQuest2.questsMobs[desc].quests = new()
			end
			if not tomQuest2.questsMobs[desc].quests[uid] then
				tomQuest2.questsMobs[desc].numQuests = tomQuest2.questsMobs[desc].numQuests + 1
				tomQuest2.questsMobs[desc].quests[uid] = objective
			end
		end
	end
	tomQuest2.questsObjectives[uid][objective].got[sender] = format("%s~%d", got, nGot)
	tomQuest2:computeQuestObjectiveStatus(uid, objective)
	if sender == tomQuest2.playerName then
		tomQuest2.cbh:Fire("addPlayerQuestObjective", first, "o", uid, objective, desc, qType, got, need, nGot, nNeed)
	else
		tomQuest2.cbh:Fire("addPartyQuestObjective", nil, nil, uid, objective)
	end
end

function tomQuest2:questObjectiveProgress(sender, first, uid, objective, got, nGot)
	tomQuest2.questsObjectives[uid][objective].got[sender] = format("%s~%d", got, nGot)
	tomQuest2:computeQuestObjectiveStatus(uid, objective)
	if sender == tomQuest2.playerName then
		tomQuest2.cbh:Fire("playerQuestObjectiveProgress", first, "p", uid, objective, got, nGot)
	else
		tomQuest2.cbh:Fire("partyQuestObjectiveProgress", sender, uid, objective, got, nGot)
	end
end

function tomQuest2:questLost(sender, first, uid, abandoning, message)
	local title = tomQuest2.quests[uid].title
	tomQuest2.quests[uid].playersOnQuest = tomQuest2.quests[uid].playersOnQuest - 1
	if tomQuest2.quests[uid].numObjectives > 0 then
		for objective, qoData in pairs(tomQuest2.questsObjectives[uid]) do
			tomQuest2.questsObjectives[uid][objective].got[sender] = nil
			if tomQuest2.quests[uid].playersOnQuest == 0 then
				if qoData.qType == "object" then
					if tomQuest2.questsItems[qoData.desc] then
						tomQuest2.questsItems[qoData.desc].quests[uid] = nil
						tomQuest2.questsItems[qoData.desc].numQuests = tomQuest2.questsItems[qoData.desc].numQuests - 1
						if tomQuest2.questsItems[qoData.desc].mobDrop then
							for mobName in pairs(tomQuest2.questsItems[qoData.desc].mobDrop) do
								tomQuest2.questsMobsItems[mobName].items[qoData.desc][uid] = nil
							end
						end
						if tomQuest2.questsItems[qoData.desc].numQuests == 0 then
							if tomQuest2.questsItems[qoData.desc].mobDrop then
								for mobName in pairs(tomQuest2.questsItems[qoData.desc].mobDrop) do
									tomQuest2.questsMobsItems[mobName].items[qoData.desc] = del(tomQuest2.questsMobsItems[mobName].items[qoData.desc])
									tomQuest2.questsMobsItems[mobName].numItems = tomQuest2.questsMobsItems[mobName].numItems - 1
									if tomQuest2.questsMobsItems[mobName].numItems == 0 then
										tomQuest2.questsMobsItems[mobName] = del(tomQuest2.questsMobsItems[mobName])
									end
								end
							end
							tomQuest2.questsItems[qoData.desc] = del(tomQuest2.questsItems[qoData.desc])
						end
					end
				elseif qoData.qType == "monster" then
					tomQuest2.questsMobs[qoData.desc].quests[uid] = nil
					tomQuest2.questsMobs[qoData.desc].numQuests = tomQuest2.questsMobs[qoData.desc].numQuests - 1
					if tomQuest2.questsMobs[qoData.desc].numQuests == 0 then
						tomQuest2.questsMobs[qoData.desc] = del(tomQuest2.questsMobs[qoData.desc])
					end
				end
				tomQuest2.questsObjectives[uid][objective] = del(tomQuest2.questsObjectives[uid][objective])
			else
				tomQuest2:computeQuestObjectiveStatus(uid, objective)
			end
		end
	end
	tomQuest2.quests[uid].complete[sender] = nil
	if tomQuest2.quests[uid].playersOnQuest == 0 then
		if sender == tomQuest2.playerName then
			if currentUidMacro == uid then
				tomQuest2:generateSpecialItemMacro()
			end
			tomQuest2.questsTitleToUid[tomQuest2.quests[uid].title] = nil
			tomQuest2.questsIdToUid[tomQuest2.quests[uid].id] = nil
			if tomQuest2.quests[uid].tracked == 1 then
				tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ = tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ - 1
				tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ_PlQC = tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ_PlQC - tomQuest2.quests[uid].playerComplete
				tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ_PaQC = tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ_PaQC - tomQuest2.quests[uid].partyComplete
				tomQuest2.numTQ_PlQC = tomQuest2.numTQ_PlQC - tomQuest2.quests[uid].playerComplete
				tomQuest2.numTQ_PaQC = tomQuest2.numTQ_PaQC - tomQuest2.quests[uid].partyComplete
				if tomQuest2.quests[uid].weightedLevel - tomQuest2.playerLevel <= tonumber(prdb.quests.maxLevelDiff) then
					tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ_BML = tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ_BML - 1
					tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ_PlQC_BML = tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ_PlQC_BML - tomQuest2.quests[uid].playerComplete
					tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ_PaQC_BML = tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ_PaQC_BML - tomQuest2.quests[uid].partyComplete
					tomQuest2.numTQ_PlQC_BML = tomQuest2.numTQ_PlQC_BML - tomQuest2.quests[uid].playerComplete
					tomQuest2.numTQ_PaQC_BML = tomQuest2.numTQ_PaQC_BML - tomQuest2.quests[uid].partyComplete
				end
			end
			tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numQuests = tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numQuests - 1
			tomQuest2.questsByZone[tomQuest2.quests[uid].zone].quests[uid] = nil
			if tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numQuests == 0 then
				tomQuest2.questsByZone[tomQuest2.quests[uid].zone] = del(tomQuest2.questsByZone[tomQuest2.quests[uid].zone])
			end
		else
			if tomQuest2.quests[uid].tracked == 1 then
				tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numTQ = tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numTQ - 1
				tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numTQ_PaQC = tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numTQ_PaQC - tomQuest2.quests[uid].partyComplete
				tomQuest2.numTQ_PaQC = tomQuest2.numTQ_PaQC - tomQuest2.quests[uid].partyComplete
				if tomQuest2.quests[uid].weightedLevel - tomQuest2.playerLevel <= tonumber(prdb.quests.maxLevelDiff) then
					tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numTQ_BML = tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numTQ_BML - 1
					tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numTQ_PaQC_BML = tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numTQ_PaQC_BML - tomQuest2.quests[uid].partyComplete
					tomQuest2.numTQ_PaQC_BML = tomQuest2.numTQ_PaQC_BML - tomQuest2.quests[uid].partyComplete
				end
			end
			tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numQuests = tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numQuests - 1
			tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].quests[uid] = nil
			if tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numQuests == 0 then
				tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]] = del(tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]])
			end
		end
		if tomQuest2.quests[uid].numObjectives > 0 then
			tomQuest2.questsObjectives[uid] = del(tomQuest2.questsObjectives[uid])
		end
		tomQuest2.quests[uid] = del(tomQuest2.quests[uid])
	else
		if sender == tomQuest2.playerName then
			if currentUidMacro == uid then
				tomQuest2:generateSpecialItemMacro()
			end
			tomQuest2.questsTitleToUid[tomQuest2.quests[uid].title] = nil
			tomQuest2.questsIdToUid[tomQuest2.quests[uid].id] = nil
			tomQuest2.quests[uid].id = nil
			tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numQuests = tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numQuests - 1
			tomQuest2.questsByZone[tomQuest2.quests[uid].zone].quests[uid] = nil
			if not tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]] then
				tomQuest2:addZone(LC["PARTYQUESTSZONE"])
			end
			tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numQuests = tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numQuests + 1
			tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].quests[uid] = true
			if tomQuest2.quests[uid].tracked == 1 then
				tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ_PlQC = tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ_PlQC - tomQuest2.quests[uid].playerComplete
				tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ_PaQC = tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ_PaQC - tomQuest2.quests[uid].partyComplete
				tomQuest2.numTQ_PlQC = tomQuest2.numTQ_PlQC - tomQuest2.quests[uid].playerComplete
				if tomQuest2.quests[uid].weightedLevel - tomQuest2.playerLevel <= tonumber(prdb.quests.maxLevelDiff) then
					tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ_BML = tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ_BML - 1
					tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ_PlQC_BML = tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ_PlQC_BML - tomQuest2.quests[uid].playerComplete
					tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ_PaQC_BML = tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ_PaQC_BML - tomQuest2.quests[uid].partyComplete
					tomQuest2.numTQ_PlQC_BML = tomQuest2.numTQ_PlQC_BML - tomQuest2.quests[uid].playerComplete
					tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numTQ_BML = tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numTQ_BML + 1
					tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numTQ_PaQC_BML = tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numTQ_PaQC_BML + tomQuest2.quests[uid].partyComplete
				end
				tomQuest2.quests[uid].playerComplete = 0
				tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ = tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numTQ - 1
				tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numTQ_PaQC = tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numTQ_PaQC + tomQuest2.quests[uid].partyComplete
				tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numTQ = tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numTQ + 1
			end
			if tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numQuests == 0 then
				tomQuest2.questsByZone[tomQuest2.quests[uid].zone] = del(tomQuest2.questsByZone[tomQuest2.quests[uid].zone])
			end
		end
		tomQuest2:computeQuestStatus(uid)
	end
	if sender == tomQuest2.playerName then
		tomQuest2.cbh:Fire("playerQuestLost", first, "l", uid, abandoning)
	elseif message == 1 then
		tomQuest2.cbh:Fire("partyQuestLost", sender, uid, abandoning, title)
	end
end

function tomQuest2:computeQuestLevel()
	for uid, questData in pairs(tomQuest2.quests) do
		if questData.shortTag == ELITE and prdb.quests.eliteAddLevel > 0 then
			tomQuest2.quests[uid].weightedLevel = (questData.level or 0) + prdb.quests.eliteAddLevel
		elseif questData.shortTag == GROUP and prdb.quests.groupAddLevel > 0 then
			tomQuest2.quests[uid].weightedLevel = (questData.level or 0) + (prdb.quests.groupAddLevel * (tomQuest2.quests[uid].group - (tomQuest2.numPartyMembers or (GetNumPartyMembers() or 0))))
		elseif questData.shortTag == HEROIC and prdb.quests.heroicAddLevel > 0 then
			tomQuest2.quests[uid].weightedLevel = (questData.level or 0) + (prdb.quests.heroicAddLevel * (5 - (tomQuest2.numPartyMembers or (GetNumPartyMembers() or 0))))
		elseif questData.shortTag == DUNGEON and prdb.quests.dungeonAddLevel > 0 then
			tomQuest2.quests[uid].weightedLevel = (questData.level or 0) + (prdb.quests.dungeonAddLevel * (5 - (tomQuest2.numPartyMembers or (GetNumPartyMembers() or 0))))
		else
			tomQuest2.quests[uid].weightedLevel = (questData.level or 0)
		end
	end
	tomQuest2.cbh:Fire("questsWeightedLevelUpdated")
end

function tomQuest2:computeQuestTag()
	for uid, questData in pairs(tomQuest2.quests) do
		if questData.level > 0 or questData.shortTag ~= "" or questData.group > 0 or questData.daily ~= "" then
			tomQuest2.quests[uid].tag = format("[%s%s%s%s]", questData.level or "", questData.shortTag ~= "" and prdb.quests[tags[questData.shortTag]] or "", questData.group > 0 and questData.group or "", questData.daily and (questData.shortTag ~= DAILY and prdb.quests[tags[DAILY]]))
		else
			tomQuest2.quests[uid].tag = ""
		end
	end
	tomQuest2.cbh:Fire("updateQuestsTracker")
end

------------------------
--- QUEST_LOG_UPDATE ---
------------------------
function tomQuest2:QUEST_LOG_UPDATE(event)
	if (not event or event ~= "FORCED") and questLogUpdatePending then return end
	questLogUpdatePending = false
	questLogNeedUpdate = false
	for i in pairs(tomQuest2.questsIdToUid) do tomQuest2.questsIdToUid[i] = nil end
	for i in pairs(tomQuest2.questsTitleToUid) do tomQuest2.questsTitleToUid[i] = nil end
	local selectionId = GetQuestLogSelection()
	local timeStamp = GetTime()
	tomQuest2.numCompletedQuests = 0
	tomQuest2.numEntries, tomQuest2.numQuests = GetNumQuestLogEntries()
	tomQuest2.numCompletedDailyQuests = GetDailyQuestsCompleted()
	if tomQuest2.numEntries > 0 then
		local zone
		for id = 1, tomQuest2.numEntries do
			local title, level, tag, group, header, collapsed, complete, daily = GetQuestLogTitle(id)
			if header then
				zone = title
				if collapsed then
					collapsedHeader[zone] = true
				elseif collapsedHeader[zone] then
					collapsedHeader[zone] = nil
				end
				if not tomQuest2.questsByZone[zone] then
					tomQuest2:addZone(zone)
				end
			else
				local uid = tonumber(match(GetQuestLink(id), 'quest:(%d+)'))
				tomQuest2.questsIdToUid[id] = uid
				tomQuest2.questsTitleToUid[title] = uid
				if not tomQuest2.quests[uid] or (tomQuest2.quests[uid] and not tomQuest2.quests[uid].complete[tomQuest2.playerName]) then
					local numObjectives = tomQuest2.quests[uid] and tomQuest2.quests[uid].numObjectives or GetNumQuestLeaderBoards(id)
					tomQuest2:addQuest(tomQuest2.playerName, firstRun, uid, title, level or 0, tag or "", group or 0, daily or 0, numObjectives or 0, complete or 0, zone, id)
					tomQuest2.quests[uid].id = id
				else
					tomQuest2.quests[uid].id = id
					if tomQuest2.quests[uid].complete[tomQuest2.playerName] ~= complete then
						if complete == 1 then
							tomQuest2:questComplete(tomQuest2.playerName, firstRun, uid)
							tomQuest2.numCompletedQuests = tomQuest2.numCompletedQuests + 1
						elseif complete == -1 then
							tomQuest2:questFailed(tomQuest2.playerName, firstRun, uid)
						end
					elseif complete == 1 then
						tomQuest2.numCompletedQuests = tomQuest2.numCompletedQuests + 1
					end
				end
				local started = 0
				local numObjectivesOK = 0
				if tomQuest2.quests[uid].numObjectives > 0 then
					for objective = 1, tomQuest2.quests[uid].numObjectives do
						local need, got, nNeed, nGot, desc
						local description, qType, done = GetQuestLogLeaderBoard(objective, id)
						local unknownQT = nil
						if qType == "item" or qType == "object" then
							qType = "object"
							desc, got, need = description:match(objectsPattern)
							nNeed, nGot = tonumber(need), tonumber(got)
						elseif qType == "player" then
							desc, got, need = description:match(playersPattern)
							nNeed, nGot = tonumber(need), tonumber(got)
						elseif qType == "monster" then
							desc, got, need = description:match(monstersPattern)
							if desc == nil or got == nil or need == nil then
								desc, got, need = description:match(objectsPattern)
							end
							nNeed, nGot = tonumber(need), tonumber(got)
						elseif qType == "reputation" then
							desc, got, need = description:match(factionPattern)
							nNeed, nGot = reactions[need], reactions[got]
						elseif qType == "event" then
							 desc = description
							 got = done and 1 or 0
							 need = 1
							 nNeed, nGot = need, got
						else
							DEFAULT_CHAT_FRAME:AddMessage("Quest objective type unknown please report to Gagou on the WowAce, WowInterface or in game on the EU Drek'Thar Server")
							DEFAULT_CHAT_FRAME:AddMessage(qType)
							DEFAULT_CHAT_FRAME:AddMessage(description)
							unknownQT = true
						end
						if not desc or strtrim(desc) == "" then
							questLogNeedUpdate = true
							questLogUpdatePending = true
						elseif not unknownQT then
							numObjectivesOK = numObjectivesOK + 1
							if not tomQuest2.questsObjectives[uid][objective] or not tomQuest2.questsObjectives[uid][objective].got[tomQuest2.playerName] then
								tomQuest2:addQuestObjective(tomQuest2.playerName, firstRun, uid, objective, desc, qType, got, need, nGot, nNeed)
								if nGot ~= 0 then
									started = 1
								end
							elseif tomQuest2.questsObjectives[uid][objective].got[tomQuest2.playerName] then
								local _, oldNGot = tomQuest2.questsObjectives[uid][objective].got[tomQuest2.playerName]:match("^(.-)~(%d-)$")
								oldNGot = tonumber(oldNGot)
								if oldNGot ~= nGot then
									tomQuest2:questObjectiveProgress(tomQuest2.playerName, firstRun, uid, objective, got, nGot)
								end
							end
						end
					end
				end
				tomQuest2.quests[uid].timeStamp = timeStamp
				if not tomQuest2.quests[uid].sync and tomQuest2.quests[uid].numObjectives == numObjectivesOK then
					tomQuest2.quests[uid].sync = true
					tomQuest2.cbh:Fire("endPlayerQuest", firstRun, "r", uid, started)
				end
			end
		end
	end
	if GetQuestLogSelection() ~= selectionId then
		SelectQuestLogEntry(selectionId)
	end
	if not firstRun then
		for uid, qData in pairs(tomQuest2.quests) do
			if qData.complete[tomQuest2.playerName] and qData.timeStamp ~= timeStamp and not collapsedHeader[qData.zone] then
				tomQuest2:questLost(tomQuest2.playerName, false, uid, abandoning, 0)
			elseif qData.complete[tomQuest2.playerName] and qData.timeStamp ~= timeStamp and collapsedHeader[qData.zone] then
				tomQuest2.quests[uid].id = nil
			end
		end
	else
		tomQuest2.cbh:Fire("updateQuestsTracker")
		tomQuest2.cbh:Fire("initParty")
	end
	if questLogNeedUpdate then
		tomQuest2:ScheduleTimer("QUEST_LOG_UPDATE", 0.1, "FORCED")
	end
	tomQuest2.cbh:Fire("questLogUpdated", true, true)
	firstRun = false
	abandoning = nil
end

-----------------------
--- PLAYER_LEVEL_UP ---
-----------------------
function tomQuest2:PLAYER_LEVEL_UP(event, newPlayerLevel)
	tomQuest2.playerLevel = newPlayerLevel
	for zone, zData in pairs(tomQuest2.questsByZone) do
		if zData.minLevel ~= 0 then
			local r, g, b = LibTourist:GetLevelColor(zone)
			tomQuest2.questsByZone[zone].color = format("%02x%02x%02x", r*255, g*255, b*255)
		end
	end
	for uid, qData in pairs(tomQuest2.quests) do
		local qc = tomQuest2:GetDifficultyColor(qData.level - tomQuest2.playerLevel)
		tomQuest2.quests[uid].color = format("%02x%02x%02x", qc.r*255, qc.g*255, qc.b*255)
	end
	tomQuest2.cbh:Fire("questLogUpdated", false, true)
	tomQuest2.cbh:Fire("questsWeightedLevelUpdated")
end

--------------------------------
--- displayQuestInformations ---
--------------------------------
function tomQuest2:displayQuestInformations(uid)
	if prdb.modules["informations"] then
		tomQuest2.cbh:Fire("displayQuestInformations", uid)
	elseif not prdb.modules["informations"] and tomQuest2.quests[uid].complete[tomQuest2.playerName] and tomQuest2.quests[uid].id then
		if QuestLogFrame:IsVisible() then
			if GetQuestLogSelection() == tomQuest2.quests[uid].id then
				return HideUIPanel(QuestLogFrame)
			end
		else
			ShowUIPanel(QuestLogFrame)
		end
		if (tomQuest2.numEntries > QUESTS_DISPLAYED) then
			if (tomQuest2.quests[uid].id <= tomQuest2.numEntries - QUESTS_DISPLAYED) then
				FauxScrollFrame_SetOffset(QuestLogListScrollFrame, tomQuest2.quests[uid].id - 1)
				QuestLogListScrollFrameScrollBar:SetValue((tomQuest2.quests[uid].id - 1) * QUESTLOG_QUEST_HEIGHT)
			else
				FauxScrollFrame_SetOffset(QuestLogListScrollFrame, tomQuest2.numEntries - QUESTS_DISPLAYED)
				QuestLogListScrollFrameScrollBar:SetValue((tomQuest2.numEntries - QUESTS_DISPLAYED) * QUESTLOG_QUEST_HEIGHT)
			end
		end
		SelectQuestLogEntry(tomQuest2.quests[uid].id)
		QuestLog_SetSelection(tomQuest2.quests[uid].id)
		_G["QuestLogTitle"..(tomQuest2.quests[uid].id - FauxScrollFrame_GetOffset(QuestLogListScrollFrame))]:Click()
	end
end

------------------------
--- Macro Generation ---
------------------------
function tomQuest2:generateSpecialItemMacro(uid)
	if uid and tomQuest2.quests[uid].id and tomQuest2.quests[uid].specialItemLink then
		currentUidMacro = uid
		if InCombatLockdown() then return end
		local found, _, itemString = find(tomQuest2.quests[uid].specialItemLink, "^|c%x+|H(.+)|h%[.*%]")
		local index = GetMacroIndexByName("tomQuest2_SI")
		if index == 0 then
			index = CreateMacro("tomQuest2_SI", 1, "#showtooltip\n/use "..itemString, 0)
			PickupMacro(index)
		else
			EditMacro(index, "tomQuest2_SI", 1, "#showtooltip\n/use "..itemString)
		end
	elseif not uid then
		currentUidMacro = nil
		local index = GetMacroIndexByName("tomQuest2_SI")
		if index == 0 then
			index = CreateMacro("tomQuest2_SI", 1, "", 0)
			PickupMacro(index)
		else
			EditMacro(index, "tomQuest2_SI", 1, "")
		end
	end
end

--------------------------
--- Keybinding globals ---
--------------------------
BINDING_HEADER_TOMQUEST2 = "tomQuest2"