-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("dailyTooltip", "AceEvent-3.0")

-----------------------
--- Library loading ---
-----------------------
local LDT = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-DailyTooltip")
local dataobj = LibStub("LibDataBroker-1.1"):NewDataObject("tomQuest2Daily", {type = "data source", icon = "Interface\\AddOns\\tomQuest2\\media\\icon"})

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local pairs = pairs
local format = string.format
local mod = mod
local GetTime = GetTime
local ToggleFrame = ToggleFrame
local CreateFrame = CreateFrame
local GetQuestResetTime = GetQuestResetTime
local GetMaxDailyQuests = GetMaxDailyQuests

-----------------
--- variables ---
-----------------
module.modName = LDT["Daily Quests Tooltip"]
local prdb, options
local dailytip
local counter = 0
local delay = 60
local maxDaily = GetMaxDailyQuests()
local dailytimer = CreateFrame("Frame")
local fd = 1/86400
local fh = 1/3600
local fm = 1/60

-----------------------
--- Default profile ---
-----------------------
local defaults = {
	profile = {
		showNumCompletedDailyQuests = true,
		showMaxDailyQuest = true,
		showRemainingTime = true,
	}
}

---------------
--- Options ---
---------------
local optGetter, optSetter
do
	function optGetter(info)
		return prdb[info[#info]]
	end
	function optSetter(info, value)
		prdb[info[#info]] = value
		module:updateLDBText()
	end
end

local function getOptions()
	if not options then
		options = {
			type = "group",
			name = module.modName,
			desc = module.modName,
			get = optGetter,
			set = optSetter,
			args = {
				showNumCompletedDailyQuests = {
					name = LDT["Show number of completed daily quest"],
					type = "toggle",
					width = "full",
					order = 11,
				},
				showMaxDailyQuest = {
					name = LDT["Show maximum daily quests number"],
					type = "toggle",
					width = "full",
					order = 12,
				},
				showRemainingTime = {
					name = LDT["Show remaining time before end of day"],
					type = "toggle",
					width = "full",
					order = 13,
				},
			},
		}
	end
	return options
end

----------------------
--- Initialization ---
----------------------
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("dailyTooltip", defaults)
	prdb = self.db.profile
	self:SetEnabledState(core:GetModuleEnabled("dailyTooltip"))
	core:RegisterModuleOptions("dailyTooltip", getOptions, module.modName)
end

function module:OnEnable()
	core.RegisterCallback(self, "updateQuestsTracker", "updateLDBText")
	module:updateLDBText()
end

function dailytimer:OnUpdate(elapsed)
	counter = counter + elapsed
	if counter >= delay then
		module:updateLDBText()
	end
end

function module:OnDisable()
	core.UnregisterAllCallbacks(self)
	self:Refresh()
end

function module:Refresh()
	prdb = self.db.profile
end

---------------------------
--- Main Module Program ---
---------------------------
function module:updateLDBText(event)
	if prdb.showNumCompletedDailyQuests then
		if prdb.showMaxDailyQuest then
			if prdb.showRemainingTime then
				local remainingTime = GetQuestResetTime()
				dataobj.text = format("|cff%s%s/%s|r |cff%s%d:%02d|r", core:getColorFromCompletionHEX(core.numCompletedDailyQuests/maxDaily), core.numCompletedDailyQuests, maxDaily, core:getColorFromCompletionHEX(remainingTime*fd), remainingTime*fh, mod(remainingTime*fm, 60))
			else
				dataobj.text = format("|cff%s%s/%s|r", core:getColorFromCompletionHEX(core.numCompletedDailyQuests/maxDaily), core.numCompletedDailyQuests, maxDaily)
			end
		elseif prdb.showRemainingTime then
			local remainingTime = GetQuestResetTime()
			dataobj.text = format("|cff%s%s|r |cff%s%d:%02d|r", core:getColorFromCompletionHEX(core.numCompletedDailyQuests/maxDaily), core.numCompletedDailyQuests, core:getColorFromCompletionHEX(remainingTime*fd), remainingTime*fh, mod(remainingTime*fm, 60))
		else
			dataobj.text = format("|cff%s%s|r", core:getColorFromCompletionHEX(core.numCompletedDailyQuests/maxDaily), core.numCompletedDailyQuests)
		end
	elseif prdb.showMaxDailyQuest then
		if prdb.showRemainingTime then
			local remainingTime = GetQuestResetTime()
			dataobj.text = format("|cffffffff%s|r |cff%s%d:%02d|r", maxDaily, core:getColorFromCompletionHEX(remainingTime*fd), remainingTime*fh, mod(remainingTime*fm, 60))
		else
			dataobj.text = maxDaily
		end	
	elseif prdb.showRemainingTime then
		local remainingTime = GetQuestResetTime()
		dataobj.text = format("|cff%s%d:%02d|r", core:getColorFromCompletionHEX(remainingTime*fd), remainingTime*fh, mod(remainingTime*fm, 60))
	else
		dataobj.text = ""
	end
end

function dataobj.OnClick(self,button)
	if button == "RightButton" then
		InterfaceOptionsFrame_OpenToCategory(core.optionsFrames.Profiles)
		InterfaceOptionsFrame_OpenToCategory(core.optionsFrames.tomQuest2)
	elseif IsShiftKeyDown() then
		tQ2toggleQuestTracker()
	elseif IsAltKeyDown() then
		tQ2toggleAchievementTracker()
	else
		ToggleFrame(QuestLogFrame)
	end
end

dailytimer:SetScript("OnUpdate", dailytimer.OnUpdate)