-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("questsTracker", "AceEvent-3.0", "AceHook-3.0")

-----------------------
--- Library loading ---
-----------------------
local LQTr = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-QuestsTracker")
local LSM = LibStub("LibSharedMedia-3.0")
local LibQTip = LibStub("LibQTip-1.0")

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local pairs, unpack, format, type, match, sub = pairs, unpack, string.format, type, string.match, string.sub
local ToggleFrame = ToggleFrame
local SetAbandonQuest, GetAbandonQuestItems, GetAbandonQuestName = SetAbandonQuest, GetAbandonQuestItems, GetAbandonQuestName
local SelectQuestLogEntry = SelectQuestLogEntry
local QuestLogPushQuest = QuestLogPushQuest
local MouseIsOver = MouseIsOver
local ItemRefTooltip = ItemRefTooltip
local GetQuestLogSelection = GetQuestLogSelection
local GetQuestLink = GetQuestLink
local CreateFrame, CreateFont = CreateFrame, CreateFont
local UnitIsDead, UnitAffectingCombat = UnitIsDead, UnitAffectingCombat
local GetSubZoneText, GetZoneText = GetSubZoneText, GetZoneText
local IsShiftKeyDown, IsControlKeyDown, IsAltKeyDown = IsShiftKeyDown, IsControlKeyDown, IsAltKeyDown
local ChatEdit_InsertLink = ChatEdit_InsertLink
local ShowUIPanel, HideUIPanel = ShowUIPanel, HideUIPanel

-----------------
--- Variables ---
-----------------
module.modName = LQTr["Quests Tracker"]
local prdb, chdb, options
local subZone, currentZone, oldZone = "", "", ""
local data = {true, true, true}
local numTQ = 0
local numTQ_BML = 0
local anchorPosition = {[1] = "TOPLEFT", [2] = "TOP", [3] = "TOPRIGHT", [4] = "BOTTOMLEFT", [5] = "BOTTOM", [6] = "BOTTOMRIGHT"}
local autoTrack = {[1] = LQTr["Always"], [2] = LQTr["Below Max Level"], [3] = LQTr["Never"]}
local hideQuests = {[1] = LQTr["When you've completed the quest"], [2] = LQTr["When everyone in the party have completed the quest"], [3] = LQTr["Never"]}
local hideObjectives = {[1] = LQTr["When you've completed the objective"], [2] = LQTr["When everyone in the party have completed the objective"], [3] = LQTr["Never"]}
local outlines = {[""] = LQTr["None"], ["OUTLINE"] = LQTr["Outline"], ["THICKOUTLINE"] = LQTr["Thick Outline"]}

local tracker, trackerAnchor
local hTrFont = CreateFont("hTrFont")
local qTrFont = CreateFont("qTrFont")
local qoTrFont = CreateFont("qoTrFont")

----------------------------
--- Tracker Anchor Frame ---
----------------------------
local anchor = CreateFrame("Button", nil, UIParent)
anchor:SetHeight(24)
anchor:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16, insets = {left = 5, right = 5, top = 5, bottom = 5}, tile = true, tileSize = 16})
anchor:SetBackdropColor(0.09, 0.09, 0.19, 0.5)
anchor:SetBackdropBorderColor(0.5, 0.5, 0.5, 1)
local text = anchor:CreateFontString(nil, nil, "GameFontNormalSmall")
text:SetPoint("CENTER")
text:SetText(LQTr["Tracker"])
anchor:SetWidth(text:GetStringWidth() + 8)
anchor:SetMovable(true)
anchor:RegisterForDrag("LeftButton")
anchor:SetClampedToScreen(true)
anchor:SetScript("OnDragStart", function(self)
	self:StartMoving()
end)
anchor:SetScript("OnDragStop", function(self)
	self:StopMovingOrSizing()
	prdb.point, prdb.x, prdb.y = "BOTTOMLEFT", self:GetLeft(), self:GetBottom()
	trackerAnchor = anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	LibQTip:Release(tracker)
	tracker = nil
	module:updateQuestTracker()
end)

-----------------------
--- Default profile ---
-----------------------
local defaults = {
	profile = {
		point = "CENTER",
		x = 0,
		y = 0,
		onLogIn = false,
		onQuestAccepted = 3,
		onObjectiveProgress = true,
		onSubZone = true,
		onZone = true,
		onDaily = true,
		onClass = true,
		showAnchor = true,
		anchorPosition = 2,
		toggleTracker = true,
		hideTrackerInCombat = true,
		hideCompletedQuest = 2,
		hideCompletedObjective = 2,
		onlyCurrentZone = false,
		limitQuestLevel = false,
		headerFont = LSM.DefaultMedia.font,
		headerFontSize = 12,
		headerFontOutline = "",
		questFont = LSM.DefaultMedia.font,
		questFontSize = 10,
		questFontOutline = "",
		questObjectiveFont = LSM.DefaultMedia.font,
		questObjectiveFontSize = 8,
		questObjectiveFontOutline = "",
		maxWidth = 300,
		scale = 1,
		showQuestItemIcon = true,
		questItemIconSize = 24,
		questsCellPadding = 5,
		objectivesCellPadding = 5,
		backDropColor = {0.09, 0.09, 0.09, 1},
		borderColor = {1, 1, 1, 1},
		colorizeQuestName = true,
		defaultQuestNameColor = "ffffff",
		colorizeQuestLevel = true,
		defaultQuestLevelColor = "ffffff",
		colorizeQuestObjective = true,
		defaultQuestObjectiveColor = "ffffff",
		colorizeQuestObjectiveCounter = true,
		defaultQuestObjectiveCounterColor = "ffffff",
		displayHeader = true,
		autoCollapseHeader = true,
		collapseButtonSize = 24,
		colorizeHeaderName = true,
		defaultHeaderNameColor = "ffffff",
		hideZoneLevelRange = false,
		colorizeHeaderLevel = true,
		defaultHeaderLevelColor = "ffffff",
		chatLink  = true,
		questsObjectives = true,
		questsMacro = true,
		clickTooltip = true,
		shareQuest = true,
		trackQuest = true,
		abandonQuest = true,
		questLog = true,
		overTooltip = true,
		cqiObjectiveStartNPC = true,
		cqiObjectiveEndNPC = true,
		cqiObjective = true,
		mmObjective = true,
	},
}

---------------
--- Options ---
---------------
local optGetter, optSetter, colorGetter, colorSetter, hexColorGetter, hexColorSetter, anchorSetter, combatSetter, maxLDSetter, headFontSetter, questFontSetter, qoFontSetter, hideHeader, hideHeader2, hideHeader3, hideHeader4, hideLimitQuestLevel, hideItemIcon, hideQNColor, hideQLColor, hideQOColor, hideCQI, hideMM
do
	function optGetter(info)
		return prdb[info[#info]]
	end
	function optSetter(info, value)
		prdb[info[#info]] = value
		LibQTip:Release(tracker)
		tracker = nil
		module:updateQuestTracker()
	end
	function colorGetter(info)
		return unpack(prdb[info[#info]])
	end
	function colorSetter(info, arg1, arg2, arg3, arg4)
		prdb[info[#info]][1] = arg1
		prdb[info[#info]][2] = arg2
		prdb[info[#info]][3] = arg3
		prdb[info[#info]][4] = arg4
		LibQTip:Release(tracker)
		tracker = nil
		module:updateQuestTracker()
	end
	function hexColorGetter(info)
		return tonumber(sub(prdb[info[#info]], 1, 2), 16) / 255, tonumber(sub(prdb[info[#info]], 3, 4), 16) / 255, tonumber(sub(prdb[info[#info]], 5, 6), 16) / 255
	end
	function hexColorSetter(info, arg1, arg2, arg3)
		prdb[info[#info]] = format("%02x%02x%02x", arg1*255, arg2*255, arg3*255)
		LibQTip:Release(tracker)
		tracker = nil
		module:updateQuestTracker()
	end
	function anchorSetter(info, value)
		prdb[info[#info]] = value
		if value then
			anchor:Show()
		else
			anchor:Hide()
		end
		LibQTip:Release(tracker)
		tracker = nil
		module:updateQuestTracker()
	end
	function combatSetter(info, value)
		prdb[info[#info]] = value
		if value then
			module:RegisterEvent("PLAYER_REGEN_ENABLED", "updateQuestsTracker")
			module:RegisterEvent("PLAYER_REGEN_DISABLED", "hideQuestsTracker")
		else
			module:UnregisterEvent("PLAYER_REGEN_ENABLED")
			module:UnregisterEvent("PLAYER_REGEN_DISABLED")
		end
	end
	function maxLDGetter(info)
		return core.db.profile.quests[info[#info]]
	end
	function maxLDSetter(info, value)
		core.db.profile.quests[info[#info]] = value
		module:questsWeightedLevelUpdated()
		LibQTip:Release(tracker)
		tracker = nil
		module:updateQuestTracker()
	end
	function headFontSetter(info, value)
		prdb[info[#info]] = value
		hTrFont:SetFont(LSM:Fetch("font", prdb.headerFont), prdb.headerFontSize, prdb.headerFontOutline)
		LibQTip:Release(tracker)
		tracker = nil
		module:updateQuestTracker()
	end
	function questFontSetter(info, value)
		prdb[info[#info]] = value
		qTrFont:SetFont(LSM:Fetch("font", prdb.questFont), prdb.questFontSize, prdb.questFontOutline)
		LibQTip:Release(tracker)
		tracker = nil
		module:updateQuestTracker()
	end
	function qoFontSetter(info, value)
		prdb[info[#info]] = value
		qoTrFont:SetFont(LSM:Fetch("font", prdb.questObjectiveFont), prdb.questObjectiveFontSize, prdb.questObjectiveFontOutline)
		LibQTip:Release(tracker)
		tracker = nil
		module:updateQuestTracker()
	end
	function hideHeader()
		if prdb.displayHeader then
			return false
		else
			return true
		end
	end
	function hideHeader2()
		if (not prdb.displayHeader) or (prdb.displayHeader and prdb.colorizeHeaderName) then
			return true
		else
			return false
		end
	end
	function hideHeader3()
		if (not prdb.displayHeader) or (prdb.displayHeader and prdb.hideZoneLevelRange) then
			return true
		else
			return false
		end
	end
	function hideHeader4()
		if (not prdb.displayHeader) or (prdb.displayHeader and (prdb.hideZoneLevelRange or (not prdb.hideZoneLevelRange and prdb.colorizeHeaderLevel))) then
			return true
		else
			return false
		end
	end
	function hideLimitQuestLevel()
		if prdb.limitQuestLevel then
			return false
		else
			return true
		end
	end
	function hideItemIcon()
		if prdb.showQuestItemIcon then
			return false
		else
			return true
		end
	end
	function hideQNColor()
		if not prdb.colorizeQuestName then
			return false
		else
			return true
		end
	end
	function hideQLColor()
		if not prdb.colorizeQuestLevel then
			return false
		else
			return true
		end
	end
	function hideQOColor()
		if not prdb.colorizeQuestObjective then
			return false
		else
			return true
		end
	end
	function hideCQI()
		if Cartographer_QuestInfo then
			return false
		else
			return true
		end
	end
	function hideMM()
		if MobMapDotParentFrame then
			return false
		else
			return true
		end
	end
end

local function getOptions()
	if not options then
		options = {
			type = "group",
			name = module.modName,
			desc = module.modName,
			get = optGetter,
			set = optSetter,
			args = {
				onLogIn = {
					name = LQTr["Track all quests on log in"],
					type = "toggle",
					width = "full",
					order = 0,
				},
				onQuestAccepted = {
					name = LQTr["Auto track quests when you accept them"],
					type = "select",
					values = autoTrack,
					width = "full",
					order = 1,
				},
				onObjectiveProgress = {
					name = LQTr["Auto track quests on objective progress"],
					type = "toggle",
					width = "full",
					order = 2,
				},
				onSubZone = {
					name = LQTr["Auto track quests on sub zone change"],
					type = "toggle",
					width = "full",
					order = 3,
				},
				onZone = {
					name = LQTr["Auto track quests on zone change"],
					type = "toggle",
					width = "full",
					order = 4,
				},
				onDaily = {
					name = LQTr["Auto track daily quests"],
					type = "toggle",
					width = "full",
					order = 5,
				},
				onClass = {
					name = LQTr["Auto track class quests"],
					type = "toggle",
					width = "full",
					order = 6,
				},
				showAnchor = {
					name = LQTr["Show quest tracker anchor"],
					type = "toggle",
					width = "full",
					set = anchorSetter,
					order = 7,
				},
				anchorPosition = {
					name = LQTr["Anchor position"],
					type = "select",
					values = anchorPosition,
					width = "full",
					order = 8,
				},
				trackerHeader = {
					type = "header",
					name = LQTr["Tracker Settings"],
					order = 12,
				},
				displayHeader = {
					name = LQTr["Display quests by zone"],
					type = "toggle",
					width = "full",
					order = 13,
				},
				autoCollapseHeader = {
					name = LQTr["Automatically expand /collapse zone header on zone change"],
					hidden = hideHeader,
					type = "toggle",
					width = "full",
					order = 14,
				},
				toggleTracker = {
					name = LQTr["Display tracker"],
					type = "toggle",
					width = "full",
					order = 15,
				},
				hideTrackerInCombat = {
					name = LQTr["Hide quest tracker in combat"],
					type = "toggle",
					set = combatSetter,
					width = "full",
					order = 16,
				},
				hideCompletedQuest = {
					name = LQTr["Hide completed quest"],
					type = "select",
					values = hideQuests,
					width = "full",
					order = 17,
				},
				hideCompletedObjective = {
					name = LQTr["Hide completed quests Objectives"],
					type = "select",
					values = hideObjectives,
					width = "full",
					order = 18,
				},
				onlyCurrentZone = {
					name = LQTr["Hide quests that are not in the current zone"],
					type = "toggle",
					width = "full",
					order = 19,
				},
				limitQuestLevel = {
					name = LQTr["Hide high level quests (You'll have to choose the maximum level difference)"],
					type = "toggle",
					width = "full",
					order = 20,
				},
				maxLevelDiff = {
					name = LQTr["Maximum level difference between the weighted quest level and your level"],
					hidden = hideLimitQuestLevel,
					type = "range",
					min = 0,
					max = 10,
					step = 1,
					get = maxLDGetter,
					set = maxLDSetter,
					width = "full",
					order = 21,
				},
				headerFont = {
					name = LQTr["Zone header font"],
					hidden = hideHeader,
					type = "select",
					dialogControl = 'LSM30_Font',
					values = AceGUIWidgetLSMlists.font,
					set = headFontSetter,
					width = "full",
					order = 22,
				},
				headerFontSize = {
					name = LQTr["Zone header font size"],
					hidden = hideHeader,
					type = "range",
					min = 6,
					max = 20,
					step = 1,
					set = headFontSetter,
					width = "full",
					order = 23,
				},
				headerFontOutline = {
					name = LQTr["Zone header font outline"],
					type = "select",
					values = outlines,
					set = headFontSetter,
					width = "full",
					order = 24,
				},
				questFont = {
					name = LQTr["Quests font"],
					type = "select",
					dialogControl = 'LSM30_Font',
					values = AceGUIWidgetLSMlists.font,
					set = questFontSetter,
					width = "full",
					order = 25,
				},
				questFontSize = {
					name = LQTr["Quests font size"],
					type = "range",
					min = 6,
					max = 20,
					step = 1,
					set = questFontSetter,
					width = "full",
					order = 26,
				},
				questFontOutline = {
					name = LQTr["Quests font outline"],
					type = "select",
					values = outlines,
					set = questFontSetter,
					width = "full",
					order = 27,
				},
				questObjectiveFont = {
					name = LQTr["Quests objectives font"],
					type = "select",
					dialogControl = 'LSM30_Font',
					values = AceGUIWidgetLSMlists.font,
					set = qoFontSetter,
					width = "full",
					order = 28,
				},
				questObjectiveFontSize = { 
					name = LQTr["Quests objectives font size"],
					type = "range",
					min = 6,
					max = 20,
					step = 1,
					set = qoFontSetter,
					width = "full",
					order = 29,
				},
				questObjectiveFontOutline = {
					name = LQTr["Quests objectives font outline"],
					type = "select",
					values = outlines,
					set = qoFontSetter,
					width = "full",
					order = 30,
				},
				maxWidth = {
					name = LQTr["Quests Tracker column max width"],
					type = "range",
					min = 50,
					max = 600,
					step = 5,
					width = "full",
					order = 31,
				},
				scale = {
					name = LQTr["Quests Tracker scale"],
					type = "range",
					min = 0,
					max = 1,
					step = 0.1,
					width = "full",
					order = 32,
				},
				showQuestItemIcon = {
					name = LQTr["Show special quest item icon behind quest name"],
					type = "toggle",
					width = "full",
					order = 33,
				},
				questItemIconSize = {
					name = LQTr["Quest item icon size"],
					type = "range",
					hidden = hideItemIcon,
					min = 10,
					max = 40,
					step = 1,
					width = "full",
					order = 34,
				},
				questsCellPadding = {
					name = LQTr["Quests cell padding"],
					hidden = hideHeader,
					type = "range",
					min = 0,
					max = 50,
					step = 1,
					width = "full",
					order = 35,
				},
				objectivesCellPadding = {
					name = LQTr["Quests objectives cell padding"],
					type = "range",
					min = 0,
					max = 50,
					step = 1,
					width = "full",
					order = 36,
				},
				collapseButtonSize = {
					name = LQTr["Collapse Button Size"],
					hidden = hideHeader,
					type = "range",
					min = 10,
					max = 40,
					step = 1,
					width = "full",
					order = 37,
				},
				backDropColor = {
					name = LQTr["Quests Tracker background color"],
					type = "color",
					hasAlpha = true,
					get = colorGetter,
					set = colorSetter,
					width = "full",
					order = 38,
				},
				borderColor = {
					name = LQTr["Quests Tracker border color"],
					type = "color",
					hasAlpha = true,
					get = colorGetter,
					set = colorSetter,
					width = "full",
					order = 39,
				},
				colorizeHeaderName = {
					name = LQTr["Colorize Quests Header Name"],
					hidden = hideHeader,
					type = "toggle",
					width = "full",
					order = 40,
				},
				defaultHeaderNameColor = {
					name = LQTr["Default header name color"],
					type = "color",
					hasAlpha = false,
					hidden = hideHeader2,
					get = hexColorGetter,
					set = hexColorSetter,
					width = "full",
					order = 41,
				},
				hideZoneLevelRange = {
					name = LQTr["Hide zone level range"],
					hidden = hideHeader,
					type = "toggle",
					width = "full",
					order = 42,
				},
				colorizeHeaderLevel = {
					name = LQTr["Colorize Quests Header Level"],
					hidden = hideHeader3,
					type = "toggle",
					width = "full",
					order = 43,
				},
				defaultHeaderLevelColor = {
					name = LQTr["Default header level color"],
					type = "color",
					hidden = hideHeader4,
					hasAlpha = false,
					get = hexColorGetter,
					set = hexColorSetter,
					width = "full",
					order = 44,
				},
				colorizeQuestName = {
					name = LQTr["Colorize Quests Name"],
					type = "toggle",
					width = "full",
					order = 45,
				},
				defaultQuestNameColor = {
					name = LQTr["Default quest name color"],
					type = "color",
					hidden = hideQNColor,
					hasAlpha = false,
					get = hexColorGetter,
					set = hexColorSetter,
					width = "full",
					order = 46,
				},
				colorizeQuestLevel = {
					name = LQTr["Colorize Quests Level"],
					type = "toggle",
					width = "full",
					order = 47,
				},
				defaultQuestLevelColor = {
					name = LQTr["Default quest level color"],
					type = "color",
					hidden = hideQLColor,
					hasAlpha = false,
					get = hexColorGetter,
					set = hexColorSetter,
					width = "full",
					order = 48,
				},
				colorizeQuestObjective = {
					name = LQTr["Colorize Quests Objectives"],
					type = "toggle",
					width = "full",
					order = 49,
				},
				defaultQuestObjectiveColor = {
					name = LQTr["Default quest objective color"],
					type = "color",
					hidden = hideQOColor,
					hasAlpha = false,
					get = hexColorGetter,
					set = hexColorSetter,
					width = "full",
					order = 50,
				},
				colorizeQuestObjectiveCounter = {
					name = LQTr["Colorize quests objectives counter"],
					type = "toggle",
					width = "full",
					order = 51,
				},
				trackerInteraction = {
					type = "header",
					name = LQTr["Tracker interaction"],
					order = 52,
				},
				chatLink  = {
					name = LQTr["Insert quest link in chat on shift right click"],
					type = "toggle",
					width = "full",
					order = 53,
				},
				questsObjectives = {
					name = LQTr["Insert quest objectives in chat on ctrl right click"],
					type = "toggle",
					width = "full",
					order = 54,
				},
				questsMacro = {
					name = LQTr["Update special quest item macro"],
					type = "toggle",
					width = "full",
					order = 55,
				},
				clickTooltip = {
					name = LQTr["Display of quest tooltip on right click"],
					type = "toggle",
					width = "full",
					order = 56,
				},
				shareQuest = {
					name = LQTr["Share quest on ctrl left click"],
					type = "toggle",
					width = "full",
					order = 57,
				},
				trackQuest = {
					name = LQTr["Track quest on shift left click"],
					type = "toggle",
					width = "full",
					order = 58,
				},
				abandonQuest = {
					name = LQTr["Abandon quest on alt left click"],
					type = "toggle",
					width = "full",
					order = 59,
				},
				questLog = {
					name = LQTr["Display quest log on left click"],
					type = "toggle",
					width = "full",
					order = 60,
				},
				overTooltip = {
					name = LQTr["Display quest tooltip on mouse over"],
					type = "toggle",
					width = "full",
					order = 61,
				},
				cqiObjectiveStartNPC = {
					name = LQTr["Display of start npc in Cartographer quest Info"],
					hidden = hideCQI,
					type = "toggle",
					width = "full",
					order = 62,
				},
				cqiObjectiveEndNPC = {
					name = LQTr["Display of end npc in Cartographer Quest Info"],
					hidden = hideCQI,
					type = "toggle",
					width = "full",
					order = 63,
				},
				cqiObjective = {
					name = LQTr["Display of objective in Cartographer Quest Info"],
					hidden = hideCQI,
					type = "toggle",
					width = "full",
					order = 64,
				},
				mmObjective = {
					name = LQTr["Display of objective in your map with MobMap"],
					hidden = hideMM,
					type = "toggle",
					width = "full",
					order = 65,
				},
			},
		}
	end
	return options
end

----------------------
--- Initialization ---
----------------------
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("questsTracker", defaults, "Default")
	prdb = self.db.profile
	chdb = self.db.char
	self:SetEnabledState(core:GetModuleEnabled("questsTracker"))
	core:RegisterModuleOptions("questsTracker", getOptions, module.modName)
	chdb.questsTracked = chdb.questsTracked or {}
	chdb.questsSubZoneAutoTracked = chdb.questsSubZoneAutoTracked or {}
	chdb.questsZoneAutoTracked = chdb.questsZoneAutoTracked or {}
	chdb.zoneDisplay = chdb.zoneDisplay or {}
end

function module:OnEnable()
	hTrFont:SetFont(LSM:Fetch("font", prdb.headerFont), prdb.headerFontSize, prdb.headerFontOutline)
	qTrFont:SetFont(LSM:Fetch("font", prdb.questFont), prdb.questFontSize, prdb.questFontOutline)
	qoTrFont:SetFont(LSM:Fetch("font", prdb.questObjectiveFont), prdb.questObjectiveFontSize, prdb.questObjectiveFontOutline)
	subZone = GetSubZoneText() or ""
	currentZone = GetZoneText() or ""
	anchor:SetPoint(prdb.point, prdb.x, prdb.y)
	if not prdb.showAnchor then anchor:Hide() else anchor:Show() end
	trackerAnchor = anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	if prdb.hideTrackerInCombat then
		self:RegisterEvent("PLAYER_REGEN_ENABLED","updateQuestsTracker")
		self:RegisterEvent("PLAYER_REGEN_DISABLED","hideQuestsTracker")
	end
	self:RegisterEvent("MINIMAP_ZONE_CHANGED", "zoneChanged")
	self:RegisterEvent("ZONE_CHANGED", "zoneChanged")
	self:RegisterEvent("ZONE_CHANGED_INDOORS", "zoneChanged")
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA", "zoneChanged")
	self:SecureHook("AddQuestWatch")
	self:SecureHook("RemoveQuestWatch")
	self:SecureHook("GetNumQuestWatches")
	core.RegisterCallback(self, "endPlayerQuest")
	core.RegisterCallback(self, "endPartyQuest")
	core.RegisterCallback(self, "playerQuestComplete", "playerQuestChanged")
	core.RegisterCallback(self, "playerQuestFailed", "playerQuestChanged")
	core.RegisterCallback(self, "partyQuestComplete", "partyQuestChanged")
	core.RegisterCallback(self, "partyQuestFailed", "partyQuestChanged")
	core.RegisterCallback(self, "playerQuestObjectiveProgress")
	core.RegisterCallback(self, "partyQuestObjectiveProgress")
	core.RegisterCallback(self, "playerQuestLost")
	core.RegisterCallback(self, "partyQuestLost")
	core.RegisterCallback(self, "toggleTrackedQuest")
	core.RegisterCallback(self, "questObjectiveAdded")
	core.RegisterCallback(self, "questDescriptionAdded")
	core.RegisterCallback(self, "updateQuestsTracker")
	core.RegisterCallback(self, "questsWeightedLevelUpdated")
end

function module:OnDisable()
	anchor:Hide()
	LibQTip:Release(tracker)
	tracker = nil
	core.UnregisterAllCallbacks(self)
	self:Refresh()
end


function module:Refresh()
	prdb = self.db.profile
	chdb = self.db.char
	if not self:IsEnabled() then return end
	chdb.questsTracked = chdb.questsTracked or {}
	chdb.questsSubZoneAutoTracked = chdb.questsSubZoneAutoTracked or {}
	chdb.questsZoneAutoTracked = chdb.questsZoneAutoTracked or {}
	chdb.zoneDisplay = chdb.zoneDisplay or {}
	hTrFont:SetFont(LSM:Fetch("font", prdb.headerFont), prdb.headerFontSize, prdb.headerFontOutline)
	qTrFont:SetFont(LSM:Fetch("font", prdb.questFont), prdb.questFontSize, prdb.questFontOutline)
	qoTrFont:SetFont(LSM:Fetch("font", prdb.questObjectiveFont), prdb.questObjectiveFontSize, prdb.questObjectiveFontOutline)
	anchor:SetPoint(prdb.point, prdb.x, prdb.y)
	if not prdb.showAnchor then anchor:Hide() else anchor:Show() end
	trackerAnchor = anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	if prdb.hideTrackerInCombat then
		self:RegisterEvent("PLAYER_REGEN_ENABLED","updateQuestsTracker")
		self:RegisterEvent("PLAYER_REGEN_DISABLED","hideQuestsTracker")
	end
end

---------------------------
--- Main Module Program ---
---------------------------
--------------
--- Events ---
--------------
function module:zoneChanged()
	if UnitIsDead("player") then return end
	local trackerNeedUpdate = false
	subZone = GetSubZoneText() or ""
	currentZone = GetZoneText() or ""
	if prdb.onZone and currentZone ~= oldZone then
		for uid in pairs(chdb.questsZoneAutoTracked) do
			if core.quests[uid] then
				trackerNeedUpdate = true
				module:RemoveQuestWatch(nil, uid, true)
			else
				chdb.questsZoneAutoTracked[uid] = nil
			end
		end
		for uid, qData in pairs(core.quests) do
			if not chdb.questsTracked[uid] and qData.zone == currentZone then
				chdb.questsZoneAutoTracked[uid] = true
				trackerNeedUpdate = true
				module:AddQuestWatch(nil, nil, uid, true)
			end
		end
	end
	if prdb.onSubZone and subZone ~= "" then
		for uid in pairs(chdb.questsSubZoneAutoTracked) do
			if core.quests[uid] then
				trackerNeedUpdate = true
				module:RemoveQuestWatch(nil, uid, true)
			else
				chdb.questsSubZoneAutoTracked[uid] = nil
			end
		end
		for uid, qData in pairs(core.quests) do
			if not chdb.questsTracked[uid] and ((qData.objective and qData.objective:find(subZone, 1, true)) or (qData.description and qData.description:find(subZone, 1, true))) then
				chdb.questsSubZoneAutoTracked[uid] = true
				trackerNeedUpdate = true
				module:AddQuestWatch(nil, nil, uid, true)
			end
		end
	end
	if prdb.displayHeader and prdb.autoCollapseHeader then
		for zone, zData in pairs(core.questsByZone) do
			if currentZone == zone then
				chdb.zoneDisplay[zone] = 1
			elseif zData.zone then
				chdb.zoneDisplay[zone] = 0
			else
				chdb.zoneDisplay[zone] = 1
			end
		end
		trackerNeedUpdate = true
	end
	if trackerNeedUpdate then
		LibQTip:Release(tracker)
		tracker = nil
		module:updateQuestTracker()
	end
	oldZone = currentZone
end
-------------
--- Hooks ---
-------------
function module:AddQuestWatch(id, timeToTrack, uid, batch)
	if not uid then
		uid = core.questsIdToUid[id]
		if not uid then return end
	end
	if (not chdb.questsTracked[uid]) or (chdb.questsTracked[uid] and core.quests[uid].tracked == 0) then
		numTQ = numTQ + 1
		core.numTQ_PlQC = core.numTQ_PlQC + core.quests[uid].playerComplete
		core.numTQ_PaQC = core.numTQ_PaQC + core.quests[uid].partyComplete
		chdb.questsTracked[uid] = true
		core.quests[uid].tracked = 1
		if core.quests[uid].complete[core.playerName] then
			core.questsByZone[core.quests[uid].zone].numTQ = core.questsByZone[core.quests[uid].zone].numTQ + 1
			core.questsByZone[core.quests[uid].zone].numTQ_PlQC = core.questsByZone[core.quests[uid].zone].numTQ_PlQC + core.quests[uid].playerComplete
			core.questsByZone[core.quests[uid].zone].numTQ_PaQC = core.questsByZone[core.quests[uid].zone].numTQ_PaQC + core.quests[uid].partyComplete
			if (core.quests[uid].weightedLevel - core.playerLevel <= tonumber(core.db.profile.quests.maxLevelDiff)) then
				core.questsByZone[core.quests[uid].zone].numTQ_BML = core.questsByZone[core.quests[uid].zone].numTQ_BML + 1
				core.questsByZone[core.quests[uid].zone].numTQ_PlQC_BML = core.questsByZone[core.quests[uid].zone].numTQ_PlQC_BML + core.quests[uid].playerComplete
				core.questsByZone[core.quests[uid].zone].numTQ_PaQC_BML = core.questsByZone[core.quests[uid].zone].numTQ_PaQC_BML + core.quests[uid].partyComplete
				core.numTQ_PlQC_BML = core.numTQ_PlQC_BML + 1
				core.numTQ_PaQC_BML = core.numTQ_PaQC_BML + 1
			end
		else
			core.questsByZone[LQTr["PARTYQUESTSZONE"]].numTQ = core.questsByZone[LQTr["PARTYQUESTSZONE"]].numTQ + 1
			core.questsByZone[LQTr["PARTYQUESTSZONE"]].numTQ_PaQC = core.questsByZone[LQTr["PARTYQUESTSZONE"]].numTQ_PaQC + core.quests[uid].partyComplete
			if (core.quests[uid].weightedLevel - core.playerLevel <= tonumber(core.db.profile.quests.maxLevelDiff)) then
				core.questsByZone[LQTr["PARTYQUESTSZONE"]].numTQ_BML = core.questsByZone[LQTr["PARTYQUESTSZONE"]].numTQ_BML + 1
				core.questsByZone[LQTr["PARTYQUESTSZONE"]].numTQ_PaQC_BML = core.questsByZone[LQTr["PARTYQUESTSZONE"]].numTQ_PaQC_BML + core.quests[uid].partyComplete
				core.numTQ_PlQC_BML = core.numTQ_PlQC_BML + 1
				core.numTQ_PaQC_BML = core.numTQ_PaQC_BML + 1
			end
		end
		if not batch then
			LibQTip:Release(tracker)
			tracker = nil
			module:updateQuestTracker()
		end
	end
	if QuestLogFrame:IsVisible() then
		QuestLog_Update()
	end
end

function module:RemoveQuestWatch(id, uid, batch)
	if not uid then
		uid = core.questsIdToUid[id]
		if not uid then return end
	end
	if chdb.questsTracked[uid] then
		numTQ = numTQ - 1
		core.numTQ_PlQC = core.numTQ_PlQC - core.quests[uid].playerComplete
		core.numTQ_PaQC = core.numTQ_PaQC - core.quests[uid].partyComplete
		chdb.questsTracked[uid] = nil
		chdb.questsSubZoneAutoTracked[uid] = nil
		chdb.questsZoneAutoTracked[uid] = nil
		core.quests[uid].tracked = 0
		if core.quests[uid].complete[core.playerName] then
			core.questsByZone[core.quests[uid].zone].numTQ = core.questsByZone[core.quests[uid].zone].numTQ - 1
			core.questsByZone[core.quests[uid].zone].numTQ_PlQC = core.questsByZone[core.quests[uid].zone].numTQ_PlQC - core.quests[uid].playerComplete
			core.questsByZone[core.quests[uid].zone].numTQ_PaQC = core.questsByZone[core.quests[uid].zone].numTQ_PaQC - core.quests[uid].partyComplete
			if (core.quests[uid].weightedLevel - core.playerLevel <= tonumber(core.db.profile.quests.maxLevelDiff)) then
				core.questsByZone[core.quests[uid].zone].numTQ_BML = core.questsByZone[core.quests[uid].zone].numTQ_BML - 1
				core.questsByZone[core.quests[uid].zone].numTQ_PlQC_BML = core.questsByZone[core.quests[uid].zone].numTQ_PlQC_BML - core.quests[uid].playerComplete
				core.questsByZone[core.quests[uid].zone].numTQ_PaQC_BML = core.questsByZone[core.quests[uid].zone].numTQ_PaQC_BML - core.quests[uid].partyComplete
				core.numTQ_PlQC_BML = core.numTQ_PlQC_BML - 1
				core.numTQ_PaQC_BML = core.numTQ_PaQC_BML - 1
			end
		else
			core.questsByZone[LQTr["PARTYQUESTSZONE"]].numTQ = core.questsByZone[LQTr["PARTYQUESTSZONE"]].numTQ - 1
			core.questsByZone[LQTr["PARTYQUESTSZONE"]].numTQ_PaQC = core.questsByZone[LQTr["PARTYQUESTSZONE"]].numTQ_PaQC - core.quests[uid].partyComplete
			if (core.quests[uid].weightedLevel - core.playerLevel <= tonumber(core.db.profile.quests.maxLevelDiff)) then
				core.questsByZone[LQTr["PARTYQUESTSZONE"]].numTQ_BML = core.questsByZone[LQTr["PARTYQUESTSZONE"]].numTQ_BML - 1
				core.questsByZone[LQTr["PARTYQUESTSZONE"]].numTQ_PaQC_BML = core.questsByZone[LQTr["PARTYQUESTSZONE"]].numTQ_PaQC_BML - core.quests[uid].partyComplete
				core.numTQ_PlQC_BML = core.numTQ_PlQC_BML - 1
				core.numTQ_PaQC_BML = core.numTQ_PaQC_BML - 1
			end
		end
		if not batch then
			LibQTip:Release(tracker)
			tracker = nil
			module:updateQuestTracker()
		end
	end
	if QuestLogFrame:IsVisible() then
		QuestLog_Update()
	end
end

function module:GetNumQuestWatches()
	return numTQ
end

function module:isQuestTracked(uid)
	if chdb.questsTracked[uid] then
		return true
	else
		return false
	end
end

-----------------
--- Callbacks ---
-----------------
function module:questsWeightedLevelUpdated(event)
	numTQ_BML = 0
	core.numTQ_PlQC_BML = 0
	core.numTQ_PaQC_BML = 0
	for zone, zData in pairs(core.questsByZone) do
		core.questsByZone[zone].numTQ_BML = 0
		core.questsByZone[zone].numTQ_PlQC_BML = 0
		core.questsByZone[zone].numTQ_PaQC_BML = 0
		for uid in pairs(zData.quests) do
			if core.quests[uid].tracked == 1 and (core.quests[uid].weightedLevel - core.playerLevel) <= tonumber(core.db.profile.quests.maxLevelDiff) then
				numTQ_BML = numTQ_BML + 1
				core.numTQ_PlQC_BML = core.numTQ_PlQC_BML + core.quests[uid].playerComplete
				core.numTQ_PaQC_BML = core.numTQ_PaQC_BML + core.quests[uid].partyComplete
				core.questsByZone[zone].numTQ_BML = core.questsByZone[zone].numTQ_BML + 1
				core.questsByZone[zone].numTQ_PlQC_BML = core.questsByZone[zone].numTQ_PlQC_BML + core.quests[uid].playerComplete
				core.questsByZone[zone].numTQ_PaQC_BML = core.questsByZone[zone].numTQ_PaQC_BML + core.quests[uid].partyComplete
			end
		end
	end
	LibQTip:Release(tracker)
	tracker = nil
	module:updateQuestTracker()
end

function module:endPlayerQuest(event, first, commHandler, uid, started)
	local questAddedToTracker = 0
	if first then
		if chdb.questsTracked[uid] and core.quests[uid].tracked == 0 then
			AddQuestWatch(core.quests[uid].id, nil, uid, first)
		elseif prdb.onLogIn and core.quests[uid].tracked == 0 then
			AddQuestWatch(core.quests[uid].id, nil, uid, first)
		end
	elseif prdb.onQuestAccepted == 1 or (prdb.onQuestAccepted == 2 and core.playerLevel ~= MAX_PLAYER_LEVEL) then
		if not chdb.questsTracked[uid] then
			questAddedToTracker = 1
			AddQuestWatch(core.quests[uid].id, nil, uid, first)
		else
			chdb.questsSubZoneAutoTracked[uid] = nil
			chdb.questsZoneAutoTracked[uid] = nil
		end
	else
		if not chdb.questsTracked[uid] then
			if (prdb.onObjectiveProgress and started == 1) or (prdb.onDaily and core.quests[uid].daily == 1) or (prdb.onClass and core.quests[uid].class) or (prdb.onZone and core.quests[uid].zone == currentZone) then
				questAddedToTracker = 1
				AddQuestWatch(core.quests[uid].id, nil, uid, first)
			elseif prdb.onSubZone and subZone ~= "" and (core.quests[uid].objective:find(subZone, 1, true) or core.quests[uid].description:find(subZone, 1, true)) then
				chdb.questsSubZoneAutoTracked[uid] = true
				questAddedToTracker = 1
				AddQuestWatch(core.quests[uid].id, nil, uid, first)
			end
		elseif (prdb.onObjectiveProgress and started == 1) or (prdb.onDaily and core.quests[uid].tag == DAILY) or (prdb.onClass and core.quests[uid].class) or (prdb.onZone and core.quests[uid].zone == currentZone) then
			chdb.questsSubZoneAutoTracked[uid] = nil
			chdb.questsZoneAutoTracked[uid] = nil
		end
	end
	if not first and questAddedToTracker == 0 and chdb.questsTracked[uid] then
		LibQTip:Release(tracker)
		tracker = nil
		module:updateQuestTracker()
	end
end

function module:endPartyQuest(event, sender, uid, started, batch)
	local questAddedToTracker = 0
	if batch then
		if chdb.questsTracked[uid] and core.quests[uid].tracked == 0 then
			AddQuestWatch(core.quests[uid].id, nil, uid, batch)
		end
	elseif prdb.onQuestAccepted == 1 or (prdb.onQuestAccepted == 2 and core.playerLevel ~= MAX_PLAYER_LEVEL) then
		if not chdb.questsTracked[uid] then
			questAddedToTracker = 1
			AddQuestWatch(core.quests[uid].id, nil, uid, batch)
		else
			chdb.questsSubZoneAutoTracked[uid] = nil
			chdb.questsZoneAutoTracked[uid] = nil
		end
	else
		if not chdb.questsTracked[uid] and (prdb.onObjectiveProgress and started == 1) or (prdb.onDaily and core.quests[uid].tag == DAILY) then
			questAddedToTracker = 1
			AddQuestWatch(core.quests[uid].id, nil, uid, batch)
		elseif (prdb.onObjectiveProgress and started == 1) or (prdb.onDaily and core.quests[uid].tag == DAILY) then
			chdb.questsSubZoneAutoTracked[uid] = nil
			chdb.questsZoneAutoTracked[uid] = nil
		end
	end
	if not batch and questAddedToTracker == 0 and chdb.questsTracked[uid] then
		LibQTip:Release(tracker)
		tracker = nil
		module:updateQuestTracker()
	end
end

function module:playerQuestChanged(event, first, commHandler, uid)
	if chdb.questsTracked[uid] then
		LibQTip:Release(tracker)
		tracker = nil
		module:updateQuestTracker()
	end
end

function module:partyQuestChanged(event, sender, uid)
	if chdb.questsTracked[uid] then
		LibQTip:Release(tracker)
		tracker = nil
		module:updateQuestTracker()
	end
end

function module:playerQuestObjectiveProgress(event, first, commHandler, uid, objective, got, nGot)
	if prdb.onObjectiveProgress then
		chdb.questsSubZoneAutoTracked[uid] = nil
		chdb.questsZoneAutoTracked[uid] = nil
		if not chdb.questsTracked[uid] then
			AddQuestWatch(core.quests[uid].id, nil, uid, nil)
		else
			LibQTip:Release(tracker)
			tracker = nil
			module:updateQuestTracker()
		end
	elseif chdb.questsTracked[uid] then
		LibQTip:Release(tracker)
		tracker = nil
		module:updateQuestTracker()
	end
end

function module:partyQuestObjectiveProgress(event, sender, uid, objective, got, nGot)
	if prdb.onObjectiveProgress then
		chdb.questsSubZoneAutoTracked[uid] = nil
		chdb.questsZoneAutoTracked[uid] = nil
		if not chdb.questsTracked[uid] then
			AddQuestWatch(core.quests[uid].id, nil, uid, nil)
		else
			LibQTip:Release(tracker)
			tracker = nil
			module:updateQuestTracker()
		end
	elseif chdb.questsTracked[uid] then
		LibQTip:Release(tracker)
		tracker = nil
		module:updateQuestTracker()
	end
end

function module:playerQuestLost(event, first, commHandler, uid)
	if not core.quests[uid] and chdb.questsTracked[uid] then
		chdb.questsTracked[uid] = nil
		chdb.questsSubZoneAutoTracked[uid] = nil
		chdb.questsZoneAutoTracked[uid] = nil
		numTQ = numTQ - 1
		LibQTip:Release(tracker)
		tracker = nil
		module:updateQuestTracker()
	elseif core.quests[uid] and chdb.questsTracked[uid] then
		LibQTip:Release(tracker)
		tracker = nil
		module:updateQuestTracker()
	end
end

function module:partyQuestLost(event, sender, uid, title)
	if not core.quests[uid] and chdb.questsTracked[uid] then
		chdb.questsTracked[uid] = nil
		chdb.questsSubZoneAutoTracked[uid] = nil
		chdb.questsZoneAutoTracked[uid] = nil
		numTQ = numTQ - 1
		LibQTip:Release(tracker)
		tracker = nil
		module:updateQuestTracker()
	elseif core.quests[uid] and chdb.questsTracked[uid] then
		LibQTip:Release(tracker)
		tracker = nil
		module:updateQuestTracker()
	end
end

function module:toggleTrackedQuest(event, uid)
	if chdb.questsTracked[uid] then
		RemoveQuestWatch(core.quests[uid].id, uid, nil)
	elseif not chdb.questsTracked[uid] then
		AddQuestWatch(core.quests[uid].id, nil, uid, nil)
	end
end

function module:questObjectiveAdded(event, uid)
	if prdb.onSubZone and subZone ~= "" then
		if not chdb.questsTracked[uid] and core.quests[uid].objective and core.quests[uid].objective:find(subZone, 1, true) then
			chdb.questsSubZoneAutoTracked[uid] = true
			AddQuestWatch(core.quests[uid].id, nil, uid, nil)
		elseif chdb.questsTracked[uid] then
			LibQTip:Release(tracker)
			tracker = nil
			module:updateQuestTracker()
		end
	elseif chdb.questsTracked[uid] then
		LibQTip:Release(tracker)
		tracker = nil
		module:updateQuestTracker()
	end
end

function module:questDescriptionAdded(event, uid)
	if prdb.onSubZone and subZone ~= "" then
		if not chdb.questsTracked[uid] and core.quests[uid].description	and core.quests[uid].description:find(subZone, 1, true) then
			chdb.questsSubZoneAutoTracked[uid] = true
			AddQuestWatch(core.quests[uid].id, nil, uid, nil)
		elseif chdb.questsTracked[uid] then
			LibQTip:Release(tracker)
			tracker = nil
			module:updateQuestTracker()
		end
	elseif chdb.questsTracked[uid] then
		LibQTip:Release(tracker)
		tracker = nil
		module:updateQuestTracker()
	end
end

function module:updateQuestsTracker(event)
	LibQTip:Release(tracker)
	tracker = nil
	module:updateQuestTracker()
end

function module:hideQuestsTracker(event)
	LibQTip:Release(tracker)
	tracker = nil
end

----------------------------
--- Custom Cell Provider ---
----------------------------
local cProvider, cPrototype, basecCellPrototype = LibQTip:CreateCellProvider(LibQTip.LabelProvider)
function cPrototype:InitializeCell()
	self.fontString = self:CreateFontString()
	self.fontString:SetFontObject(GameTooltipText)
end

function cPrototype:SetupCell(tooltip, value, justification, font, bType, bValue, bValue2, maxWidth, indent, anchor)
	self:EnableMouse(true)
	self:SetScript("OnMouseDown", function(frame, button)
		if bType == "zone" then
			if chdb.zoneDisplay[bValue] == 1 then
				chdb.zoneDisplay[bValue] = 0
			else
				chdb.zoneDisplay[bValue] = 1
			end
			tracker:SetScript("OnLeave", nil)
			LibQTip:Release(tracker)
			tracker = nil
			module:updateQuestTracker()
		elseif bType == "quest" then
			if button == "RightButton" then
				if IsShiftKeyDown() then
					if core.quests[bValue].complete[core.playerName] and core.quests[bValue].id and prdb.chatLink then
						ChatFrameEditBox:Show()
						ChatEdit_InsertLink(GetQuestLink(core.quests[bValue].id))
					end
				elseif IsControlKeyDown() then
					if core.quests[bValue].numObjectives > 0 and core.quests[bValue].complete[core.playerName] and prdb.questsObjectives then
						ChatFrameEditBox:Show()
						for objective, qoData in pairs(core.questsObjectives[bValue]) do
							local got, _ = qoData:match("^(.-)~(%d-)$")
							ChatFrameEditBox:Insert(format("{%s %s/%s} ", qoData.desc, got, qoData.need))
						end
					end
				elseif IsAltKeyDown() and bType == "quest" then
					core:generateSpecialItemMacro(bValue)
				elseif core.quests[bValue].complete[core.playerName] and core.quests[bValue].id and prdb.clickTooltip then
					ItemRefTooltip:ClearLines()
					ShowUIPanel(ItemRefTooltip)
					if not ItemRefTooltip:IsVisible() then
						ItemRefTooltip:SetOwner(UIParent, "ANCHOR_PRESERVE")
					end
					local link = GetQuestLink(core.quests[bValue].id)
					if link then
						ItemRefTooltip:SetHyperlink(link)
					end
				end
			elseif button == "LeftButton" then
				if IsControlKeyDown() then
					if core.numPartyMembers > 0 and core.quests[bValue].complete[core.playerName] and core.quests[bValue].id and core.quests[bValue].sharable and prdb.shareQuest then
						local wasSelected = GetQuestLogSelection()
						SelectQuestLogEntry(core.quests[bValue].id)
						QuestLogPushQuest()
						SelectQuestLogEntry(wasSelected)
					end
				elseif IsShiftKeyDown() then
					if prdb.trackQuest then
						module:toggleTrackedQuest(nil, bValue)
					end
				elseif IsAltKeyDown() then
					if core.quests[bValue].complete[core.playerName] and core.quests[bValue].id and prdb.abandonQuest then
						local wasSelected = GetQuestLogSelection()
						SelectQuestLogEntry(core.quests[bValue].id)
						SetAbandonQuest()
						local items = GetAbandonQuestItems()
						if items then
							StaticPopup_Hide("ABANDON_QUEST")
							StaticPopup_Show("ABANDON_QUEST_WITH_ITEMS", GetAbandonQuestName(), items)
						else
							StaticPopup_Hide("ABANDON_QUEST_WITH_ITEMS")
							StaticPopup_Show("ABANDON_QUEST", GetAbandonQuestName())
						end
						SelectQuestLogEntry(wasSelected)
					end
				else
					core:displayQuestInformations(bValue)
				end
			elseif button == "MiddleButton" then
				if bValue2 ~= nil then
					core.cbh:Fire("toggleMap", bValue, bValue2)
					if Cartographer_QuestInfo then
						local q = Cartographer_QuestInfo:GetQuest(bValue)
						if not q then return end
						if IsControlKeyDown() and prdb.cqiObjectiveStartNPC then
							if q.start_npc then
								data.quest = q.title_full
								data.obj = nil
								data.npc = q.start_npc
								Cartographer_QuestInfo:OnButtonClick(nil, "start", "LeftButton", data)
							end
						elseif IsShiftKeyDown() and prdb.cqiObjectiveEndNPC then
							if q.end_npc then
								data.quest = q.title_full
								data.obj = nil
								data.npc = q.end_npc
								Cartographer_QuestInfo:OnButtonClick(nil, "end", "LeftButton", data)
							end
						elseif prdb.cqiObjective then
							for _, obj in pairs(q.objs) do
								if obj.title:find(core.questsObjectives[bValue][bValue2].desc, 1, true) then
									data.quest = q.title_full
									data.obj = obj
									data.npc = nil
								end
							end
							if data.obj ~= nil then
								Cartographer_QuestInfo:OnButtonClick(nil, "obj", "LeftButton", data)
							end
						end
					elseif MobMapDotParentFrame and MobMap_ParseQuestObjective and prdb.mmObjective then
						MobMap_ParseQuestObjective(core.questsObjectives[bValue][bValue2].desc)
					end
				end
			end
		end
	end)
	self:SetScript("OnEnter", function()
		if prdb.overTooltip and bType == "quest" then
			if bValue2 ~= nil and core.questsObjectives[bValue] and core.questsObjectives[bValue][bValue2].qType == "object" and core.questsItems[core.questsObjectives[bValue][bValue2].desc].mobDrop then
				GameTooltip:SetOwner(tooltip, anchor)
				GameTooltip:AddLine(LQTr["|cffffffffDrop from:|r"])
				for mobName in pairs(core.questsItems[core.questsObjectives[bValue][bValue2].desc].mobDrop) do
					local zones = ""
					for zone in pairs(core.questsMobsItems[mobName].mobLoc) do
						zones = zones == "" and zone or format("%s, %s", zones, zone)
					end
					if core.questsMobsItems[mobName].items[core.questsObjectives[bValue][bValue2].desc].dropRate then
						GameTooltip:AddDoubleLine(format(LQTr["  |cffffffff%s in|r |cff00ff00%s|r"], mobName, zones), format("|cffffffff%s%%|r", core.questsMobsItems[mobName].items[core.questsObjectives[bValue][bValue2].desc].dropRate))
					else
						GameTooltip:AddLine(format(LQTr["  |cffffffff%s in|r |cff00ff00%s|r"], mobName, zones))
					end
				end
				GameTooltip:Show()
			elseif core.quests[bValue] and core.quests[bValue].id then
				local link = GetQuestLink(core.quests[bValue].id)
				if link then
					GameTooltip:SetOwner(tooltip, anchor)
					GameTooltip:SetHyperlink(link)
				end
			end
		end
	end)
	self:SetScript("OnLeave", function()
		GameTooltip:Hide()
	end)

	indent = indent or 0
	local width, height = basecCellPrototype.SetupCell(self, tooltip, value, justification, font)
	local fs = self.fontString
	fs:SetPoint("TOPLEFT", self, "TOPLEFT", indent, 0)
	fs:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", 0, 0)
	width = width + indent
	if width > maxWidth then
		fs:SetWidth(maxWidth)
		height = fs:GetHeight()
		width = maxWidth
	end
	fs:Show()
	return width, height
end

function cPrototype:ReleaseCell() 
	self:EnableMouse(false) 
	self:ClearAllPoints()
	self:SetScript("OnEnter", nil) 
	self:SetScript("OnLeave", nil) 
	self:SetScript("OnMouseDown", nil) 
end

function module:updateQuestTracker()
	if (not prdb.toggleTracker) or (numTQ == 0) or (not prdb.onlyCurrentZone and ((prdb.hideCompletedQuest == 1 and ((prdb.limitQuestLevel and numTQ_BML == core.numTQ_PlQC_BML) or (not prdb.limitQuestLevel and numTQ == core.numTQ_PlQC))) or (prdb.hideCompletedQuest == 2 and ((prdb.limitQuestLevel and numTQ_BML == core.numTQ_PaQC_BML) or (not prdb.limitQuestLevel and numTQ == core.numTQ_PaQC))) or (prdb.hideCompletedQuest == 3 and (prdb.limitQuestLevel and core.numTQ_BML == 0)))) or (prdb.onlyCurrentZone and ((not core.questsByZone[currentZone]) or (core.questsByZone[currentZone].numTQ == 0) or (prdb.hideCompletedQuest == 1 and ((prdb.limitQuestLevel and core.questsByZone[currentZone].numTQ_BML == core.questsByZone[currentZone].numTQ_PlQC_BML) or (not prdb.limitQuestLevel and core.questsByZone[currentZone].numTQ == core.questsByZone[currentZone].numTQ_PlQC))) or (prdb.hideCompletedQuest == 2 and ((prdb.limitQuestLevel and core.questsByZone[currentZone].numTQ_BML == core.questsByZone[currentZone].numTQ_PaQC_BML) or (not prdb.limitQuestLevel and core.questsByZone[currentZone].numTQ == core.questsByZone[currentZone].numTQ_PaQC))) or (prdb.hideCompletedQuest == 3 and prdb.limitQuestLevel and core.questsByZone[currentZone].numTQ_BML == 0))) then return end
	tracker = LibQTip:Acquire("tomQuest2Tracker", 2, "LEFT", "RIGHT")
	tracker:EnableMouse(true)
	tracker:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])
	tracker:SetBackdropBorderColor(prdb.borderColor[1], prdb.borderColor[2], prdb.borderColor[3], prdb.borderColor[4])
	tracker:SetScale(prdb.scale)
	tracker:SetFrameStrata("BACKGROUND")
	local tempLine
	if prdb.displayHeader then
		for zone, zData in pairs(core.questsByZone) do
			if ((prdb.onlyCurrentZone and zone == currentZone) or not prdb.onlyCurrentZone) and ((prdb.hideCompletedQuest == 1 and ((prdb.limitQuestLevel and zData.numTQ_BML > zData.numTQ_PlQC_BML) or (not prdb.limitQuestLevel and zData.numTQ > zData.numTQ_PlQC))) or (prdb.hideCompletedQuest == 2 and ((prdb.limitQuestLevel and zData.numTQ_BML > zData.numTQ_PaQC_BML) or (not prdb.limitQuestLevel and zData.numTQ > zData.numTQ_PaQC))) or (prdb.hideCompletedQuest == 3 and ((prdb.limitQuestLevel and zData.numTQ_BML > 0) or (not prdb.limitQuestLevel and zData.numTQ > 0)))) then
				chdb.zoneDisplay[zone] = chdb.zoneDisplay[zone] or 1
				tempLine = tracker:AddLine()
				tracker:SetCell(tempLine, 1, format("|TInterface\\Buttons\\UI-%sButton-Up:%d|t |cff%s%s|r |cff%s%s|r", chdb.zoneDisplay[zone] == 0 and "Plus" or "Minus", prdb.collapseButtonSize, prdb.colorizeHeaderName and zData.color or prdb.defaultHeaderNameColor, zone, prdb.colorizeHeaderLevel and zData.color or prdb.defaultHeaderLevelColor, prdb.hideZoneLevelRange and "" or zData.level ), hTrFont, "LEFT", 2, cProvider, "zone", zone, nil, prdb.maxWidth, 0)
				if chdb.zoneDisplay[zone] == 1 then
					for uid in pairs(zData.quests) do
						if chdb.questsTracked[uid] and ((prdb.hideCompletedQuest == 3) or (prdb.hideCompletedQuest == 2 and core.quests[uid].partyComplete == 0) or (prdb.hideCompletedQuest == 1 and core.quests[uid].playerComplete == 0)) and ((prdb.limitQuestLevel and (core.quests[uid].weightedLevel - core.playerLevel <= tonumber(core.db.profile.quests.maxLevelDiff))) or not prdb.limitQuestLevel) then
							tempLine = tracker:AddLine()
							if prdb.showQuestItemIcon and core.quests[uid].specialItemIcon then
								tracker:SetCell(tempLine, 1, format("|cff%s%s|r|cff%s%s|r |T%s:%d|t %s", prdb.colorizeQuestLevel and core.quests[uid].color or prdb.defaultQuestLevelColor, core.quests[uid].tag, prdb.colorizeQuestName and core.quests[uid].color or prdb.defaultQuestNameColor, core.quests[uid].title, core.quests[uid].specialItemIcon, prdb.questItemIconSize, core.quests[uid].playersOnQuest > 1 and format("<%d>", core.quests[uid].playersOnQuest) or ""), qTrFont, "LEFT", 1, cProvider, "quest", uid, nil, prdb.maxWidth, prdb.questsCellPadding, trackerAnchor)
							else
								tracker:SetCell(tempLine, 1, format("|cff%s%s|r|cff%s%s|r %s", prdb.colorizeQuestLevel and core.quests[uid].color or prdb.defaultQuestLevelColor, core.quests[uid].tag, prdb.colorizeQuestName and core.quests[uid].color or prdb.defaultQuestNameColor, core.quests[uid].title, core.quests[uid].playersOnQuest > 1 and format("<%d>", core.quests[uid].playersOnQuest) or ""), qTrFont, "LEFT", 1, cProvider, "quest", uid, nil, prdb.maxWidth, prdb.questsCellPadding, trackerAnchor)
							end
							tracker:SetCell(tempLine, 2, core.quests[uid].pqStatus, qTrFont, "RIGHT", 1, cProvider, "quest", uid, nil, prdb.maxWidth, 0, trackerAnchor)
							if core.questsObjectives[uid] then
								for objective, qoData in pairs(core.questsObjectives[uid]) do
									if prdb.hideCompletedObjective == 3 or (prdb.hideCompletedObjective == 2 and qoData.partyComplete == 0) or (prdb.hideCompletedObjective == 1 and qoData.playerComplete == 0) then
										tempLine = tracker:AddLine()
										tracker:SetCell(tempLine, 1, format("|cff%s%s|r", prdb.colorizeQuestObjective and qoData.qoColor or prdb.defaultQuestObjectiveColor, qoData.desc), qoTrFont, "LEFT", 1, cProvider, "quest", uid, objective, prdb.maxWidth, prdb.objectivesCellPadding, trackerAnchor)
										tracker:SetCell(tempLine, 2, prdb.colorizeQuestObjectiveCounter and qoData.qoStatus or qoData.qoStatusNC, qoTrFont, "RIGHT", 1, cProvider, "quest", uid, objective, prdb.maxWidth, 0, trackerAnchor)
									end
								end
							else
								tempLine = tracker:AddLine()
								tracker:SetCell(tempLine, 1, core.quests[uid].objective, qoTrFont, "LEFT", 2, cProvider,"quest", nil, uid, prdb.maxWidth, prdb.objectivesCellPadding, trackerAnchor)
							end
						end
					end
				end
			end
		end
	else
		for uid in pairs(chdb.questsTracked) do
			if uid and core.quests[uid] then
				if ((prdb.onlyCurrentZone and core.quests[uid].zone == currentZone) or not prdb.onlyCurrentZone) and ((prdb.hideCompletedQuest == 1 and core.quests[uid].playerComplete == 0) or (prdb.hideCompletedQuest == 2 and core.quests[uid].partyComplete == 0) or (prdb.hideCompletedQuest == 3)) and ((prdb.limitQuestLevel and (core.quests[uid].weightedLevel - core.playerLevel <= tonumber(core.db.profile.quests.maxLevelDiff))) or not prdb.limitQuestLevel) then
					tempLine = tracker:AddLine()
					if prdb.showQuestItemIcon and core.quests[uid].specialItemIcon then
						tracker:SetCell(tempLine, 1, format("|cff%s%s|r|cff%s%s|r |T%s:%d|t %s", prdb.colorizeQuestLevel and core.quests[uid].color or prdb.defaultQuestLevelColor, core.quests[uid].tag, prdb.colorizeQuestName and core.quests[uid].color or prdb.defaultQuestNameColor, core.quests[uid].title, core.quests[uid].specialItemIcon, prdb.questItemIconSize, core.quests[uid].playersOnQuest > 1 and format("<%d>", core.quests[uid].playersOnQuest) or ""), qTrFont, "LEFT", 1, cProvider, "quest", uid, nil, prdb.maxWidth, 0, trackerAnchor)
					else
						tracker:SetCell(tempLine, 1, format("|cff%s%s|r|cff%s%s|r %s", prdb.colorizeQuestLevel and core.quests[uid].color or prdb.defaultQuestLevelColor, core.quests[uid].tag, prdb.colorizeQuestName and core.quests[uid].color or prdb.defaultQuestNameColor, core.quests[uid].title, core.quests[uid].playersOnQuest > 1 and format("<%d>", core.quests[uid].playersOnQuest) or ""), qTrFont, "LEFT", 1, cProvider, "quest", uid, nil, prdb.maxWidth, 0, trackerAnchor)
					end
					tracker:SetCell(tempLine, 2, core.quests[uid].pqStatus, qTrFont, "RIGHT", 1, cProvider,  "quests",uid, nil, prdb.maxWidth, 0, trackerAnchor)
					if core.questsObjectives[uid] then
						for objective, qoData in pairs(core.questsObjectives[uid]) do
							if prdb.hideCompletedObjective == 3 or (prdb.hideCompletedObjective == 2 and qoData.partyComplete == 0) or (prdb.hideCompletedObjective == 1 and qoData.playerComplete == 0) then
								tempLine = tracker:AddLine()
								tracker:SetCell(tempLine, 1, format("|cff%s%s|r", prdb.colorizeQuestObjective and qoData.qoColor or prdb.defaultQuestObjectiveColor, qoData.desc), qoTrFont, "LEFT", 1, cProvider, "quest", uid, objective, prdb.maxWidth, prdb.objectivesCellPadding, trackerAnchor)
								tracker:SetCell(tempLine, 2, prdb.colorizeQuestObjectiveCounter and qoData.qoStatus or qoData.qoStatusNC, qoTrFont, "RIGHT", 1, cProvider, "quest", uid, objective, prdb.maxWidth, 0, trackerAnchor)
							end
						end
					else
						tempLine = tracker:AddLine()
						tracker:SetCell(tempLine, 1, core.quests[uid].objective, qoTrFont, "LEFT", 2, cProvider, "quest", uid, nil, prdb.maxWidth, prdb.objectivesCellPadding, trackerAnchor)
					end
				end
			end
		end
	end
	tracker:SetClampedToScreen(true)
	tracker:ClearAllPoints()
	if prdb.anchorPosition == 1 then
		if prdb.showAnchor then
			tracker:SetPoint("TOPLEFT", anchor, "BOTTOMLEFT")
		else
			tracker:SetPoint("TOPLEFT", anchor, "BOTTOMLEFT", 0, 24)
		end
	elseif prdb.anchorPosition == 2 then
		if prdb.showAnchor then
			tracker:SetPoint("TOP", anchor, "BOTTOM")
		else
			tracker:SetPoint("TOP", anchor, "BOTTOM", 0, 24)
		end
	elseif prdb.anchorPosition == 3 then
		if prdb.showAnchor then
			tracker:SetPoint("TOPRIGHT", anchor, "BOTTOMRIGHT")
		else
			tracker:SetPoint("TOPRIGHT", anchor, "BOTTOMRIGHT", 0, 24)
		end
	elseif prdb.anchorPosition == 4 then
		if prdb.showAnchor then
			tracker:SetPoint("BOTTOMLEFT", anchor, "TOPLEFT")
		else
			tracker:SetPoint("BOTTOMLEFT", anchor, "TOPLEFT", 0, -24)
		end
	elseif prdb.anchorPosition == 5 then
		if prdb.showAnchor then
			tracker:SetPoint("BOTTOM", anchor, "TOP")
		else
			tracker:SetPoint("BOTTOM", anchor, "TOP", 0, -24)
		end
	elseif prdb.anchorPosition == 6 then
		if prdb.showAnchor then
			tracker:SetPoint("BOTTOMRIGHT", anchor, "TOPRIGHT")
		else
			tracker:SetPoint("BOTTOMRIGHT", anchor, "TOPRIGHT", 0, -24)
		end
	end
	tracker:Show()
end

--------------------------
--- Keybinding globals ---
--------------------------
BINDING_NAME_TQ2QUESTTRACKERTOGGLE = LQTr["Quests Tracker"]

function tQ2toggleQuestTracker()
	prdb.toggleTracker = not prdb.toggleTracker
	LibQTip:Release(tracker)
	tracker = nil
	module:updateQuestTracker()
end