-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("tooltips", "AceHook-3.0")

-----------------------
--- Library loading ---
-----------------------
local LT = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-Tooltip")

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local pairs = pairs
local format, ceil = string.format, math.ceil

-----------------
--- Variables ---
-----------------
module.modName = LT["Tooltips"]
local prdb, options

-----------------------
--- Default profile ---
-----------------------
local defaults = {
	profile = {
		mobTooltip = true,
		mobItemDrop = true,
		itemTooltip = true,
	},
}

---------------
--- Options ---
---------------
local optGetter, optSetter, hideQMI
do
	function optGetter(info)
		return prdb[info[#info]]
	end
	function optSetter(info, value)
		prdb[info[#info]] = value
	end
	function hideQMI()
		if core.questsMobsItems and prdb.mobTooltip then
			return false
		else
			return true
		end
	end
end

local function getOptions()
	if not options then
		options = {
			type = "group",
			name = module.modName,
			desc = module.modName,
			get = optGetter,
			set = optSetter,
			args = {
				mobTooltip = {
					name = LT["Mobs tooltip modification"],
					type = "toggle",
					width = "full",
					order = 10,
				},
				mobItemDrop = {
					name = LT["Mobs quests items drop using MobMap data"],
					hidden = hideQMI,
					type = "toggle",
					width = "full",
					order = 11,
				},
				itemTooltip = {
					name = LT["Item tooltip modification"],
					type = "toggle",
					width = "full",
					order = 12,
				},
			},
		}
	end
	return options 
end

----------------------
--- Initialization ---
----------------------
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("tooltips", defaults, "Default")
	prdb = self.db.profile
	self:SetEnabledState(core:GetModuleEnabled("tooltips"))
	core:RegisterModuleOptions("tooltips", getOptions, module.modName)
end

function module:OnEnable()
	self:HookScript(GameTooltip, "OnTooltipSetUnit")
	self:HookScript(GameTooltip, "OnTooltipSetItem")
end

function module:OnDisable()
	self:Refresh()
end

function module:Refresh()
	prdb = self.db.profile
end

---------------------------
--- Main Module Program ---
---------------------------
function module:OnTooltipSetUnit(tooltip, ...)
	if not prdb.mobTooltip then return end
	local name = tooltip:GetUnit()
	if not name then return end
	if core.questsMobs[name] then
		for uid, qmData in pairs(core.questsMobs[name].quests) do
			tooltip:AddDoubleLine(format("|cffffffff%s|r", core.quests[uid].title), core.questsObjectives[uid][qmData].qoStatus)
		end
	end
 	if not prdb.mobItemDrop or not core.questsMobsItems or not core.questsMobsItems[name] then return end
	for item, qmiData in pairs (core.questsMobsItems[name].items) do
		local nGot, nNeed = 0, 0
		if qmiData.dropRate then
			tooltip:AddLine(format(LT["|cffffffffDrop %s at %s%% chance|r"], item, qmiData.dropRate))
			for uid, qmiqData in pairs(qmiData.quests) do
				for player, qoData in pairs(core.questsObjectives[uid][qmiqData].got) do
					local _, nGot2 = qoData:match("^(.-)~(%d-)$")
					nNeed = nNeed + core.questsObjectives[uid][qmiqData].nNeed
					nGot = nGot + tonumber(nGot2)
				end
				tooltip:AddDoubleLine(format("  |cffffffff%s|r", core.quests[uid].title), core.questsObjectives[uid][qmiqData].qoStatus)
			end
			tooltip:AddDoubleLine(format(LT["|cffffffffKills needed at %s%% drop chance|r"], qmiData.dropRate), format("|cffffffff%s|r", ceil((nNeed - nGot)*100/qmiData.dropRate)))
		else
			tooltip:AddLine(format(LT["|cffffffffDrop %s|r"], item))
			for uid, qmiqData in pairs(qmiData.quests) do
				tooltip:AddDoubleLine(format("  |cffffffff%s|r", core.quests[uid].title), core.questsObjectives[uid][qmiqData].qoStatus)
			end
		end
	end
end

function module:OnTooltipSetItem(tooltip, ...)
	if not prdb.itemTooltip then return end
	local name = tooltip:GetItem()
	if not name or not core.questsItems[name] then return end
	for uid, qiData in pairs(core.questsItems[name].quests) do
		tooltip:AddDoubleLine(format("|cffffffff%s|r", core.quests[uid].title), core.questsObjectives[uid][qiData].qoStatus)
	end
end